from worker_bunch.astral_times.astral_times_config import ASTRAL_TIMES_JSONSCHEMA
from worker_bunch.service_logging import LOGGING_JSONSCHEMA
from worker_bunch.database.database_config import DATABASE_CONNECTIONS_JSONSCHEMA
from worker_bunch.mqtt.mqtt_config import MQTT_JSONSCHEMA
from worker_bunch.worker.worker_config import WORKER_INSTANCES_JSONSCHEMA


class ConfigException(Exception):
    pass


class ServiceConfKey:
    LOCALE = "locale"
    DATA_DIR = "data_directory"


SERVICE_JSONSCHEMA = {
    "type": "object",
    "properties": {
        ServiceConfKey.LOCALE: {
            "type": "string",
            "minLength": 1,
            "description": "",
        },
        ServiceConfKey.DATA_DIR: {
            "type": "string",
            "minLength": 1,
            "description": "Promoted to workers, so they store their data there. Make sure it's writable. "
                           "Relative paths refer to config file.",
        },
    },
    "additionalProperties": False,
    "required": [],
}


class MainConfKey:
    ASTRAL_TIMES = "astral_times"
    DATABASE_CONNECTIONS = "database_connections"
    LOGGING = "logging"
    MQTT_BROKER = "mqtt_broker"
    SERVICE = "service"
    WORKER_INSTANCES = "worker_instances"
    WORKER_SETTINGS = "worker_settings"
    YAML_TEMPLATES = "yaml_templates"


CONFIG_JSONSCHEMA = {
    "type": "object",
    "properties": {
        MainConfKey.ASTRAL_TIMES: ASTRAL_TIMES_JSONSCHEMA,
        MainConfKey.DATABASE_CONNECTIONS: DATABASE_CONNECTIONS_JSONSCHEMA,
        MainConfKey.LOGGING: LOGGING_JSONSCHEMA,
        MainConfKey.MQTT_BROKER: MQTT_JSONSCHEMA,
        MainConfKey.SERVICE: SERVICE_JSONSCHEMA,
        MainConfKey.WORKER_INSTANCES: WORKER_INSTANCES_JSONSCHEMA,
        MainConfKey.WORKER_SETTINGS: {
            "type": "object",
            "description": "dummy section, will be filled with worker schemes based on configuration."
        },
        MainConfKey.YAML_TEMPLATES: {
            "type": "object",
            "additionalProperties": True,
            "description": "The strict schema prevents putting objects anywhere outside of given structure. This constrains the YAML "
                           "reference feature. Therefor this section is supposed to contain arbitrary objects without a schema validation. "
                           "Also nothing within this section is used by worker-bunch directly. So put your reused YAML referenced objects "
                           "here."
        },
    },
    "additionalProperties": False,
    "required": [MainConfKey.WORKER_INSTANCES],
}



import threading
from enum import Enum
from typing import List, Optional, Set


_astral_times_lock = threading.Lock()
_astral_extended_values: Optional[Set[str]] = None


class AstralTime(Enum):

    DAWN_ASTRO = "dawn_astro"
    DAWN_NAUTICAL = "dawn_nautical"
    DAWN_CIVIL = "dawn_civil"

    SUNRISE = "sunrise"
    NOON = "noon"
    SUNSET = "sunset"

    DUSK_CIVIL = "dusk_civil"
    DUSK_NAUTICAL = "dusk_nautical"
    DUSK_ASTRO = "dusk_astro"

    MIDNIGHT = "midnight"

    @classmethod
    def values(cls):
        return [astral_time.value for astral_time in AstralTime]

    @classmethod
    def extended_values(cls) -> List[str]:
        with _astral_times_lock:
            if not _astral_extended_values:
                cls._init_extended_values()
            return list(_astral_extended_values)

    @classmethod
    def extended_value_exists(cls, value: str) -> bool:
        if not value:
            return False
        with _astral_times_lock:
            global _astral_extended_values
            if not _astral_extended_values:
                cls._init_extended_values()
            return value in _astral_extended_values

    @classmethod
    def _init_extended_values(cls):
        global _astral_extended_values
        _astral_extended_values = set()
        for astral_time in AstralTime:
            _astral_extended_values.add(astral_time.value)
        for i in range(2, 19):
            _astral_extended_values.add(f"dawn_{i}")
            _astral_extended_values.add(f"dusk_{i}")


class AstralTimesConfKey:

    LATITUDE = "latitude"
    LONGITUDE = "longitude"
    ELEVATION = "elevation"


ASTRAL_TIMES_JSONSCHEMA = {
    "additionalProperties": False,
    "required": [AstralTimesConfKey.LATITUDE, AstralTimesConfKey.LONGITUDE, AstralTimesConfKey.ELEVATION],
    "type": "object",
    "properties": {
        AstralTimesConfKey.LATITUDE: {"type": "number", "minimum": -90, "maximum": 90},
        AstralTimesConfKey.LONGITUDE: {"type": "number", "minimum": -180, "maximum": 180},
        AstralTimesConfKey.ELEVATION: {
            "type": "number",
            "minimum": -1000,
            "maximum": 9000,
            "description": "in meters",
        },
    },
    "description": "A concrete location is necessary to use the astral time service. You don't need to configured it, "
                   "but this will lead to errors in case you use it.",
}



import json
import logging
import sys
from typing import List, Optional

import click
from jsonschema import ValidationError
from yaml.parser import ParserError

from worker_bunch.astral_times.astral_times_manager import AstralTimesManager
from worker_bunch.service_config import ConfigException
from worker_bunch.service_configurator import ServiceConfigurator
from worker_bunch.service_logging import LOGGING_CHOICES, ServiceLogging
from worker_bunch.database.database_manager import DatabaseManager
from worker_bunch.dispatcher import Dispatcher
from worker_bunch.mqtt.mqtt_client import MqttClient, MqttClientFactory
from worker_bunch.mqtt.mqtt_proxy import MqttProxy
from worker_bunch.runner import Runner
from worker_bunch.utils.time_utils import TimeUtils
from worker_bunch.worker.worker import Worker, WorkerSetup
from worker_bunch.worker.worker_factory import WorkerFactory

_logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--config-file",
    help="Config file",
)
@click.option(
    "--json-schema",
    is_flag=True,
    help="Prints the config file JSON schema and exits. (JSON schema is used to validate the YAML config.)"
)
@click.option(
    "--log-file",
    help="Log file (if stated journal logging is disabled)"
)
@click.option(
    "--log-level",
    help="Log level",
    type=click.Choice(LOGGING_CHOICES, case_sensitive=False),
)
@click.option(
    "--print-log-console",
    is_flag=True,
    help="Print log output to console too."
)
@click.option(
    "--skip-log-times",
    is_flag=True,
    help="Skip log timestamp (systemd/journald logs get their own timestamp)."
)
@click.option(
    "--test-single",
    help="Test/debug/run a single worker, once, single-threaded...",
)
def worker_bunch_main(config_file, json_schema, log_file, log_level, print_log_console, skip_log_times, test_single):
    """A task/rule engine framework. It bunches a set of worker threads."""
    # noinspection SpellCheckingInspection
    config_error_code = 78  # sysexits.h: define EX_CONFIG 78 /* configuration error */

    try:
        if json_schema:
            output = generate_json_schema_info(config_file)
            print(output)
        else:
            run_service(config_file, log_file, log_level, print_log_console, skip_log_times, test_single)

    except KeyboardInterrupt:
        pass  # exits 0 by default
    except ConfigException as ex:
        _logger.error(ex)
        sys.exit(config_error_code)

    except ParserError as ex:
        _logger.error("parsing error in config file:\n%s", ex)
        sys.exit(config_error_code)
    except ValidationError as ex:
        _logger.error("error in config file (see JSON schema):\n"
                      "    json-path: %s\n"
                      "    problem:   %s\n"
                      "    validator: %s (argument(s): %s)",
                      ex.json_path, ex.message, ex.validator, ex.validator_value)
        sys.exit(config_error_code)
    except Exception as ex:
        _logger.exception(ex)
        sys.exit(1)  # a simple return is not understood by click


def _shutdown_workers(workers: List[Worker]):
    for worker in workers:
        try:
            worker.stop()
        except Exception as ex:
            _logger.exception(ex)

    time_to_wait = 10.0

    while True:
        time_to_wait -= TimeUtils.sleep(0.02)
        alive_workers = [w for w in workers if w.is_alive()]
        if not alive_workers:
            break
        if time_to_wait < 0:
            _logger.warn("workers doesn't closing properly: %s", ",".join([w.name for w in workers]))
            break


def run_service(config_file, log_file, log_level, print_log_console, skip_log_times, test_single):
    dispatcher: Optional[Dispatcher] = None
    mqtt_client: Optional[MqttClient] = None
    mqtt_proxy: Optional[MqttProxy] = None
    workers: List[Worker] = []

    try:
        # configuring
        service_config = ServiceConfigurator()
        service_config.read_config_file(config_file)

        ServiceLogging.configure(
            service_config.get_logging_config(),
            log_file, log_level, print_log_console, skip_log_times
        )
        _logger.debug("start")

        service_config.init_locale()
        service_config.init_data_dir()

        worker_instances_config = service_config.get_worker_instances_config()
        worker_dict = WorkerFactory.create_workers(worker_instances_config)
        workers_settings_declarations = WorkerFactory.extract_workers_settings_declarations(worker_dict)

        service_config.revalidate_worker_settings(workers_settings_declarations)

        if test_single:
            single_worker = worker_dict.get(test_single)
            if not single_worker:
                raise ConfigException(f"single-run ({test_single}) does not exist!")
            worker_dict = {test_single: single_worker}

        workers = list(worker_dict.values())

        # bootstrapping
        astral_time_manager = AstralTimesManager(service_config.get_astral_config())
        dispatcher = Dispatcher(astral_time_manager)
        database_manager = DatabaseManager(service_config.get_database_config())

        mqtt_config = service_config.get_mqtt_config()
        mqtt_client = MqttClientFactory.create(mqtt_config) if mqtt_config else None
        mqtt_proxy = MqttProxy(mqtt_client)

        workers_settings = service_config.get_worker_settings()

        data_dir = service_config.get_data_dir()
        for worker in workers:
            worker.setup({
                WorkerSetup.ASTRAL_TIME_MANAGER: astral_time_manager,
                WorkerSetup.BASE_DATA_DIR: data_dir,
                WorkerSetup.DATABASE_MANAGER: database_manager,
                WorkerSetup.MQTT_PROXY: mqtt_proxy,
                WorkerSetup.WORKER_SETTINGS: workers_settings.get(worker.name),
            })
            worker.set_last_will()  # before set_mqtt_proxy ("last will" depends on config)!

        # start
        runner = Runner(dispatcher, mqtt_proxy, workers)
        if test_single:
            runner.run_single()
        else:
            runner.run()

    finally:
        _logger.info("shutdown")

        if dispatcher is not None:  # first: stop
            try:
                dispatcher.close()
            except Exception as ex:
                _logger.exception(ex)

        _shutdown_workers(workers)

        if mqtt_proxy is not None:
            try:
                mqtt_proxy.close()
            except Exception as ex:
                _logger.exception(ex)

        if mqtt_client is not None:
            try:
                mqtt_client.close()
            except Exception as ex:
                _logger.exception(ex)


def generate_json_schema_info(config_file: Optional[str]) -> str:
    text_blocks = []

    def append_text(text_block):
        text_blocks.append(text_block)

    def append_schema(text_block):
        text_blocks.append(json.dumps(text_block, indent=4, sort_keys=True))

    try:
        if not config_file:
            append_text(
                "The JSON schema output is supposed to contain also the worker extra settings, but therefore it needs a specific "
                "configuration. So please provide the configuration file.\n"
            )
            append_schema(ServiceConfigurator.generate_config_file_json_schema(None))
        else:
            service_config = ServiceConfigurator()
            service_config.read_config_file(config_file, skip_file_access_check=True)

            worker_instances_config = service_config.get_worker_instances_config()
            workers = WorkerFactory.create_workers(worker_instances_config)
            workers_extra_configs = WorkerFactory.extract_workers_settings_declarations(workers)

            append_schema(ServiceConfigurator.generate_config_file_json_schema(workers_extra_configs))

    except ValidationError as ex:
        append_text(
            "The JSON schema output is supposed to contain also the worker extra settings, but that is only possible "
            "if the base settings can be read flawlessly. Actually there is a problem reading the config file, so only "
            "the base JSON schema is shown.\n\nActual problem: {}\n".format(str(ex.message))
        )
        append_schema(ServiceConfigurator.generate_config_file_json_schema(None))

    return "\n".join(text_blocks)


if __name__ == '__main__':
    worker_bunch_main()  # exit codes must be handled by click!



import datetime
import unittest

from tzlocal import get_localzone

from worker_bunch.utils.time_utils import TimeUtils


class TestTimeUtils(unittest.TestCase):

    def test_is_cron_time_minute(self):
        cron = "2 * * * *"

        t = datetime.datetime(2022, 1, 30, 10, 0, 30, tzinfo=get_localzone())

        self.assertFalse(TimeUtils.hits_cron_time(cron, t))
        self.assertFalse(TimeUtils.hits_cron_time(cron, t.replace(minute=1)))
        self.assertFalse(TimeUtils.hits_cron_time(cron, t.replace(minute=3)))
        self.assertFalse(TimeUtils.hits_cron_time(cron, t.replace(minute=4)))

        self.assertTrue(TimeUtils.hits_cron_time(cron, t.replace(minute=2)))
        self.assertTrue(TimeUtils.hits_cron_time(cron, t.replace(minute=2, second=0)))

    def test_is_cron_time_per_minute(self):
        cron = "*/5 * * * *"
        t = datetime.datetime(2022, 1, 30, 10, 0, 30, tzinfo=get_localzone())

        count = 0
        for i in range(0, 60):
            if TimeUtils.hits_cron_time(cron, t.replace(minute=i)):
                count += 1

        self.assertEqual(count, 12)

    def test_is_cron_time_hour_range(self):
        cron = "1 10-15 * * *"
        t = datetime.datetime(2022, 1, 31, 10, 1, 30, tzinfo=get_localzone())

        count = 0
        for i in range(0, 24):
            if TimeUtils.hits_cron_time(cron, t.replace(hour=i)):
                count += 1

        self.assertEqual(count, 6)

    def test_is_cron_time_hour_list(self):
        cron = "1 10,11,12,13,14,15 * * *"
        t = datetime.datetime(2022, 1, 29, 10, 1, 30, tzinfo=get_localzone())

        count = 0
        for i in range(0, 24):
            if TimeUtils.hits_cron_time(cron, t.replace(hour=i)):
                count += 1

        self.assertEqual(count, 6)

    def test_cron_every_hour(self):
        now = TimeUtils.now()
        for i in range(1, 10000):
            result_hit = TimeUtils.hits_cron_time("0 0 * * *", now)
            expected_hit = now.hour == 0 and now.minute == 0
            self.assertEqual(result_hit, expected_hit)
            now = now + datetime.timedelta(seconds=45, microseconds=434)



import copy
import datetime
import logging
from collections import namedtuple
from typing import List, Optional, Dict

import astral
import astral.sun
import attr

from worker_bunch.astral_times.astral_times_config import AstralTime, AstralTimesConfKey
from worker_bunch.service_config import ConfigException
from worker_bunch.utils.time_utils import TimeUtils


_logger = logging.getLogger(__name__)


@attr.define
class AstralParsed:
    """Internal parse result. It's a little complicated to support individual "depression" values."""

    predefined: AstralTime = None

    is_dawn: Optional[bool] = None
    is_dusk: Optional[bool] = None
    depression: Optional[float] = None

    def is_valid(self) -> bool:
        mode_counter = 0
        if self.predefined:
            mode_counter += 1
        if self.is_dawn:
            mode_counter += 1
        if self.is_dusk:
            mode_counter += 1

        if mode_counter != 1:
            return False

        if self.predefined is not None:
            return True

        if self.is_dawn is True or self.is_dusk is True:
            if self.depression is None or 1 > self.depression < 18:
                return False

        return True


Location = namedtuple("Location", ["latitude", "longitude", "elevation"])


class AstralTimesManager:
    """
    Provides astral times for the configured location (latitude, longitude, altitude).

    API: https://astral.readthedocs.io/en/stable/index.html
    """

    def __init__(self, config):
        self._observer: Optional[astral.Observer] = None
        if config:
            latitude = config[AstralTimesConfKey.LATITUDE]
            longitude = config[AstralTimesConfKey.LONGITUDE]
            altitude = config[AstralTimesConfKey.ELEVATION]
            self._observer = astral.Observer(latitude=latitude, longitude=longitude, elevation=altitude)

        self._cache_time = TimeUtils.now()
        self._cached_values: Dict[str, datetime.datetime] = {}

        self._reset_cache(self._cache_time)

    def get_location(self) -> Optional[Location]:
        """Returns configured location"""
        if not self._observer:
            return None
        return Location(latitude=self._observer.latitude, longitude=self._observer.longitude, elevation=self._observer.elevation)

    def _reset_cache(self, cache_time):
        self._cached_values = {}
        self._cache_time = self.round_time_to_hour(cache_time)

    def _get_or_calc_astral_time(self, astral_key: str, pivot_time: datetime.datetime):
        if self._cache_time != self.round_time_to_hour(pivot_time):
            self._reset_cache(pivot_time)
        astral_time = self._cached_values.get(astral_key)
        if astral_time is None:
            if not self._observer:
                raise ConfigException("Astral times are not configured - missing location!")

            parsed = self.parse_astral_time_key(astral_key)
            astral_time = self.calc_astral_time(parsed, pivot_time, self._observer)
            self._cached_values[astral_key] = astral_time

        return astral_time

    def hits(self, astral_key: str, pivot_time: Optional[datetime.datetime] = None) -> List[str]:
        if pivot_time is None:
            pivot_time = TimeUtils.now()
        pivot_time = self.round_time_to_minute(pivot_time)

        astral_time = self._get_or_calc_astral_time(astral_key, pivot_time)
        return astral_time == pivot_time

    def get_astral_time(self, astral_key: str, pivot_time: Optional[datetime.datetime] = None) -> datetime.datetime:
        if pivot_time is None:
            pivot_time = TimeUtils.now()
        pivot_time = self.round_time_to_minute(pivot_time)

        astral_time = self._get_or_calc_astral_time(astral_key, pivot_time)
        return astral_time

    def get_astral_times(self, time: Optional[datetime.datetime] = None) -> Dict[str, any]:
        if time is None:
            time = TimeUtils.now()

        data = {"timestamp": time}

        for astral_time in AstralTime:
            recipe = AstralParsed(predefined=astral_time)
            try:
                data[astral_time.value] = self.calc_astral_time(recipe, time, self._observer)
            except ValueError:
                # ValueError: Sun never reaches 18 degrees below the horizon, at this location.
                data[astral_time.value] = None

        return data

    @classmethod
    def is_valid_astral_time_key(cls, value: str):
        try:
            parse_result = cls.parse_astral_time_key(value)
            return parse_result.is_valid()
        except ValueError:
            return False

    @classmethod
    def parse_astral_time_key(cls, value: str) -> AstralParsed:
        """The astral time syntax has to checked on worker level."""

        value = value.strip().lower()
        for astral_time in AstralTime:
            if astral_time.value == value:
                return AstralParsed(predefined=astral_time)

        is_dawn = value.startswith("dawn")
        is_dusk = value.startswith("dusk")
        try:
            depression = int(value[5:], 10)
        except ValueError:
            depression = None

        if depression and 1 <= depression <= 18:
            if is_dawn:
                return AstralParsed(is_dawn=True, depression=depression)
            if is_dusk:
                return AstralParsed(is_dusk=True, depression=depression)

        raise ValueError(f"no valid AstralParse value ({value})!")

    @classmethod
    def round_time_to_minute(cls, time: datetime.datetime) -> datetime.datetime:
        time = time + datetime.timedelta(seconds=30)
        time = time.replace(second=0)
        time = time.replace(microsecond=0)
        return time

    @classmethod
    def round_time_to_hour(cls, time: datetime.datetime) -> datetime.datetime:
        time = cls.round_time_to_minute(time)
        time = time.replace(minute=0)
        return time

    @classmethod
    def calc_astral_time(cls, parsed: AstralParsed, pivot_time: datetime.datetime, observer) -> datetime.datetime:
        if not parsed.is_valid():
            raise ValueError("invalid AstralParsed!")

        parsed = copy.deepcopy(parsed)

        date = pivot_time.date()
        astral_time = None

        try:
            if parsed.predefined == AstralTime.SUNRISE:
                astral_time = astral.sun.sunrise(observer, date, pivot_time.tzinfo)
            elif parsed.predefined == AstralTime.NOON:
                astral_time = astral.sun.noon(observer, date, pivot_time.tzinfo)
            elif parsed.predefined == AstralTime.SUNSET:
                astral_time = astral.sun.sunset(observer, date, pivot_time.tzinfo)
            elif parsed.predefined == AstralTime.MIDNIGHT:
                astral_time = astral.sun.midnight(observer, date, pivot_time.tzinfo)

            if astral_time is None and parsed.depression is None:
                if parsed.predefined in [AstralTime.DAWN_CIVIL, AstralTime.DUSK_CIVIL]:
                    parsed.depression = 6
                elif parsed.predefined in [AstralTime.DAWN_ASTRO, AstralTime.DUSK_ASTRO]:
                    parsed.depression = 12
                elif parsed.predefined in [AstralTime.DAWN_NAUTICAL, AstralTime.DUSK_NAUTICAL]:
                    parsed.depression = 18

            if parsed.predefined in [AstralTime.DAWN_CIVIL, AstralTime.DAWN_ASTRO, AstralTime.DAWN_NAUTICAL]:
                parsed.is_dawn = True
            elif parsed.predefined in [AstralTime.DUSK_CIVIL, AstralTime.DUSK_ASTRO, AstralTime.DUSK_NAUTICAL]:
                parsed.is_dusk = True

            if astral_time is None and parsed.is_dawn and parsed.depression is not None:
                astral_time = astral.sun.dawn(observer, date, parsed.depression, pivot_time.tzinfo)

            if astral_time is None and parsed.is_dusk and parsed.depression is not None:
                astral_time = astral.sun.dusk(observer, date, parsed.depression, pivot_time.tzinfo)

        except ValueError as ex:
            _logger.warning(f"cannot get astral time ({parsed})! {ex}")
            return None

        # if astral_time is None, then parsed.is_valid() has failed
        astral_time = cls.round_time_to_minute(astral_time)
        return astral_time



import unittest

from worker_bunch.database.database_config import DatabaseConfKey
from worker_bunch.database.database_worker import Step


class TestStep(unittest.TestCase):

    def test_create_from_dict(self):
        """keys must match property names of `Step`"""
        config = {
            DatabaseConfKey.MQTT_LAST_WILL: "last_will_123",
            DatabaseConfKey.MQTT_OUTPUT_TYPE: "json",
            DatabaseConfKey.MQTT_RETAIN: True,
            DatabaseConfKey.MQTT_TOPIC: "topic234",
            DatabaseConfKey.REPLACEMENTS: {"a": "b", "c": "d"},
            DatabaseConfKey.SCRIPT_FILE: "file567",
            DatabaseConfKey.STATEMENT: "select *"
        }

        step = Step(**config)

        for key, value in config.items():
            self.assertTrue(hasattr(step, key))
            self.assertEqual(getattr(step, key), value)



import datetime
import json


class JsonUtils:

    @classmethod
    def _default_json_serial(cls, obj):
        """JSON serializer for objects not serializable by default json code"""
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()

        raise TypeError(f"Type '{type(obj)}' is not JSON serializable!")

    @classmethod
    def dumps(cls, data, sort_keys=True, indent=None) -> str:
        return json.dumps(data, indent=indent, sort_keys=sort_keys, default=cls._default_json_serial)



from enum import Enum
from typing import Dict, List

from attr import frozen
from paho.mqtt.client import MQTTMessage


class NotificationType(Enum):
    """means MESSAGE*"""

    ASTRAL = "ASTRAL"
    CRON = "CRON"
    JUST_STARTED = "JUST_STARTED"
    MQTT_MESSAGE = "MQTT_MESSAGE"
    SINGLE_STARTED = "SINGLE_STARTED"  # used to signal "DEBUG" to worker
    TIMER = "TIMER"


NT = NotificationType


@frozen
class Notification:

    type: NotificationType

    # contains the MQTT topic or timer key
    topic: str

    # payload is not part of key. it will be skipped if new notification for the same type a topic arrive.
    payload: str = None

    def __key(self):
        return self.type, self.topic

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return NotImplemented

    @classmethod
    def ensure_string(cls, value_in) -> str:
        if isinstance(value_in, bytes):
            return value_in.decode("utf-8")

        return value_in

    @classmethod
    def create_from_mqtt(cls, mqtt_message: MQTTMessage):
        return Notification(
            type=NotificationType.MQTT_MESSAGE,
            topic=cls.ensure_string(mqtt_message.topic),
            payload=cls.ensure_string(mqtt_message.payload),
        )

    @classmethod
    def create_mqtt(cls, topic: str, payload: str):
        return Notification(type=NotificationType.MQTT_MESSAGE, topic=topic, payload=cls.ensure_string(payload))

    @classmethod
    def create_astral(cls, topic: str):
        return Notification(type=NotificationType.ASTRAL, topic=topic)

    @classmethod
    def create_timer(cls, topic: str):
        return Notification(type=NotificationType.TIMER, topic=topic)

    @classmethod
    def create_cron(cls, topic: str):
        return Notification(type=NotificationType.CRON, topic=topic)

    @classmethod
    def find(cls, notifications: List[any], ntype: NotificationType, topic: str = None):
        for notification in notifications:
            if notification.type == ntype and (topic is None or topic == notification.topic):
                return True

        return False


class NotificationBucket:

    def __init__(self):
        self._dict: Dict[Notification, Notification] = {}

    def __str__(self):
        return '{}'.format(self._dict)

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self._dict)

    def __bool__(self):
        return bool(self._dict)

    def __len__(self):
        return len(self._dict)

    def add(self, notification: Notification):
        if notification:
            self._dict[notification] = notification

    def clear(self):
        self._dict = {}

    def get_list(self):
        return list(self._dict.values())

    def get_set(self):
        return set(self._dict.values())



from typing import Dict, List, Optional

import schedule

from worker_bunch.dispatcher import Dispatcher
from worker_bunch.notification import Notification
from worker_bunch.worker.worker import Worker, WorkerSetup


class DummyWorkerConfKey:
    MQTT_TOPICS_IN = "mqtt_topics_in"
    MQTT_TOPIC_OUT = "mqtt_topic_out"
    MQTT_LAST_WILL = "mqtt_last_will"
    MQTT_RETAIN = "mqtt_retain"


DUMMY_WORKER_JSONSCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "properties": {
        DummyWorkerConfKey.MQTT_LAST_WILL: {"type": "string", "minLength": 1, "description": "MQTT last will message"},
        DummyWorkerConfKey.MQTT_RETAIN: {"type": "boolean", "description": "Make MQTT message persistent."},
        DummyWorkerConfKey.MQTT_TOPIC_OUT: {"type": "string", "minLength": 1, "description": "MQTT topic to write to"},

        DummyWorkerConfKey.MQTT_TOPICS_IN: {
            "type": "array",
            "items": {"type": "string", "minLength": 1},
            "description": "List of MQTT topics to read from"
        }
    },
    "required": [DummyWorkerConfKey.MQTT_TOPICS_IN, DummyWorkerConfKey.MQTT_TOPIC_OUT]
}


class DummyWorker(Worker):

    def __init__(self, name: str):
        super().__init__(name)

        self._mqtt_last_will = ""
        self._mqtt_retain = False
        self._mqtt_topic_out = []
        self._mqtt_topics_in = ""

    def setup(self, props: Dict[WorkerSetup, any]):
        super().setup(props)

        self._mqtt_last_will = self._worker_settings.get(DummyWorkerConfKey.MQTT_LAST_WILL)
        self._mqtt_retain = self._worker_settings.get(DummyWorkerConfKey.MQTT_RETAIN, False)
        self._mqtt_topic_out = self._worker_settings.get(DummyWorkerConfKey.MQTT_TOPIC_OUT)
        self._mqtt_topics_in = self._worker_settings.get(DummyWorkerConfKey.MQTT_TOPICS_IN, [])

    def get_partial_settings_schema(self) -> Optional[Dict[str, any]]:
        """returns a partial JSON schema for extra setting if needed. overwrite..."""
        return DUMMY_WORKER_JSONSCHEMA

    def make_partial_settings_required(self) -> bool:
        """return True to force existence via JSON schema validation. overwrite..."""
        return True

    def set_last_will(self):
        """Set the last will at mqtt client. Explicitly triggered before the mqtt client connects."""
        if self._mqtt_topic_out and self._mqtt_last_will:
            self._mqtt_proxy.set_last_will(self._mqtt_topic_out, self._mqtt_last_will)

    def subscribe_notifications(self, dispatcher: Dispatcher):

        timer_job = schedule.every(5).to(7).seconds  # type: schedule.Job
        dispatcher.subscribe_timer(self, timer_job, "keyword-timer")

        dispatcher.subscribe_cron(self, "* * * * *", "cron-every-minute")

        dispatcher.subscribe_mqtt_topics(self, self._mqtt_topics_in, 0.2)

        self._logger.info("subscribe")

    def _work(self, notifications: List[Notification]):
        for notification in notifications:
            self._logger.info("notified: %s", notification)

            message = f"dummy message: {notification.topic}"
            self._mqtt_proxy.queue(topic=self._mqtt_topic_out, payload=message, retain=self._mqtt_retain)

    def _final_work(self):
        """
        E.g.: You may send explicitly the last will(s) here (as a normal message). Kind of:
        ```
        if self._topic and self._last_will:
            self._mqtt_proxy.publish(self._topic, self._last_will)
        ```
        """
        if self._mqtt_topic_out and self._mqtt_last_will:
            self._mqtt_proxy.queue(self._mqtt_topic_out, self._mqtt_last_will, retain=self._mqtt_retain)



import unittest

from worker_bunch.astral_times.astral_times_config import AstralTime


class TestAstralTime(unittest.TestCase):

    def test_list_astral_values(self):
        values = AstralTime.values()
        self.assertTrue(values)
        self.assertIn("sunset", values)

        extended_values = AstralTime.extended_values()
        self.assertTrue(values)
        self.assertIn("sunset", values)
        self.assertTrue(len(values) < len(extended_values))

    def test_astral_values_exist(self):
        self.assertTrue(AstralTime.extended_value_exists("sunrise"))
        self.assertFalse(AstralTime.extended_value_exists("unknown..."))



import datetime
import time
import unittest
from unittest import mock

import schedule
from paho.mqtt.client import MQTTMessage
from tzlocal import get_localzone

from worker_bunch.dispatcher import Dispatcher, DispatcherListener, TopicMatch
from worker_bunch.notification import Notification


class TestTopicMatch(unittest.TestCase):

    def test_instance_vs_class_properties(self):
        t1 = TopicMatch(topic="t1", search_pattern="s1")
        listener = mock.MagicMock(DispatcherListener)
        t1.listeners.add(listener)

        t2 = TopicMatch(topic="t2")

        self.assertEqual(t1.topic, "t1")
        self.assertEqual(t1.search_pattern, "s1")
        self.assertEqual(t1.listeners, {listener})

        self.assertEqual(t2.topic, "t2")
        self.assertEqual(t2.search_pattern, None)
        self.assertEqual(t2.listeners, set())


class TestDispatcher(unittest.TestCase):

    def setUp(self):
        self.dispatcher = Dispatcher(mock.MagicMock("AstralTimeManager"))

    def tearDown(self):
        self.dispatcher.close()

    # noinspection PyTypeChecker
    def test_get_mqtt_topics(self):
        self.dispatcher.subscribe_mqtt_topics(mock.MagicMock(DispatcherListener), ["test/a", "test/#"], 0.1)
        self.dispatcher.subscribe_mqtt_topics(mock.MagicMock(DispatcherListener), ["test/a", "test/#", "test/b"], 0.1)

        topics = self.dispatcher.get_mqtt_topics()
        topics.sort()
        self.assertEqual(topics, ["test/#", "test/a", "test/b"])

    # noinspection PyTypeChecker
    def test_mqtt_messages(self):
        listener = mock.MagicMock(DispatcherListener)
        listener.add_notifications = mock.MagicMock("add_notifications")

        self.dispatcher.subscribe_mqtt_topics(listener, ["test/a", "test/#"], 0.1)

        def c_msg(topic, payload):
            m = MQTTMessage(topic=topic)
            m.payload = payload
            return m

        messages = [c_msg(b"test/a", b"test/a.1"), c_msg(b"test/a", b"test/a.2"), c_msg(b"test/b", b"test/b.1")]
        self.dispatcher.push_mqtt_messages(messages)
        time.sleep(0.05)

        m_a4 = c_msg(b"test/a", b"test/a.4")
        m_b2 = c_msg(b"test/b", b"test/b.2")

        messages = [c_msg(b"test/a", b"test/a.3"), m_a4, m_b2]
        self.dispatcher.push_mqtt_messages(messages)
        time.sleep(0.2)

        expected = {Notification.create_from_mqtt(m_a4), Notification.create_from_mqtt(m_b2)}
        listener.add_notifications.assert_called_once_with(expected)

    def test_timer(self):
        listener = mock.MagicMock(DispatcherListener)
        listener.add_notifications = mock.MagicMock("add_notifications")

        timer_job = schedule.every(5).minutes  # type: schedule.Job
        self.dispatcher.subscribe_timer(listener, timer_job, "5-minutes")

        timer_job.next_run = datetime.datetime.now()  # mock start time
        time.sleep(0.05)

        self.dispatcher.trigger_timers()

        expected = {Notification.create_timer("5-minutes")}
        listener.add_notifications.assert_called_once_with(expected)

    # noinspection PyTypeChecker
    @mock.patch("worker_bunch.utils.time_utils.TimeUtils.now")
    def test_cron(self, mocked_now):
        time_start = datetime.datetime(2022, 1, 30, 10, 0, 0, tzinfo=get_localzone())

        self.dispatcher._last_timer_execution = time_start

        mocked_now.return_value = time_start

        listener = mock.MagicMock("listener")
        listener.add_notifications = mock.MagicMock("listener")

        self.dispatcher.subscribe_cron(listener, "* * * * *", "cron-trigger")

        self.dispatcher.trigger_timers()
        listener.add_notifications.assert_not_called()

        mocked_now.return_value = time_start + datetime.timedelta(seconds=55)
        self.dispatcher.trigger_timers()

        listener.add_notifications.assert_not_called()

        mocked_now.return_value = time_start + datetime.timedelta(seconds=65)
        self.dispatcher.trigger_timers()

        expected = {Notification.create_cron("cron-trigger")}
        listener.add_notifications.assert_called_once_with(expected)



import unittest
from datetime import datetime, timedelta, timezone
from unittest import mock

from worker_bunch.astral_times.astral_times_config import AstralTimesConfKey
from worker_bunch.astral_times.astral_times_manager import AstralTimesManager
from worker_bunch.astral_times.astral_times_publisher import AstralTimesPublisher, AstralTimesPublisherConfKey
from worker_bunch.dispatcher import Dispatcher
from worker_bunch.mqtt.mqtt_proxy import MqttProxy
from worker_bunch.notification import Notification
from worker_bunch.worker.worker import WorkerSetup


class TestAstralTimesPublisher(unittest.TestCase):

    DUMMY_CONFIG = {
        AstralTimesConfKey.LATITUDE: 51.051873,
        AstralTimesConfKey.LONGITUDE: 13.741522,
        AstralTimesConfKey.ELEVATION: 125,
    }

    SETTINGS = {
        AstralTimesPublisherConfKey.MQTT_TOPIC_OUT: AstralTimesPublisherConfKey.MQTT_TOPIC_OUT,
        AstralTimesPublisherConfKey.MQTT_LAST_WILL: AstralTimesPublisherConfKey.MQTT_LAST_WILL,
        AstralTimesPublisherConfKey.MQTT_RETAIN: True,
    }

    def setUp(self):
        self.manager = AstralTimesManager(self.DUMMY_CONFIG)
        self.dispatcher = Dispatcher(self.manager)
        self.mqtt_proxy = None  # MagicMock[MqttProxy]

        self.mqtt_proxy = mock.MagicMock(MqttProxy, autospec=True)

        self.worker = AstralTimesPublisher("test")
        self.worker.setup({
            WorkerSetup.ASTRAL_TIME_MANAGER: self.manager,
            WorkerSetup.MQTT_PROXY: self.mqtt_proxy,
            WorkerSetup.WORKER_SETTINGS: self.SETTINGS,
        })

        self.reset_mqtt_proxy()

    def reset_mqtt_proxy(self):
        self.mqtt_proxy = mock.MagicMock(MqttProxy, autospec=True)
        self.worker._mqtt_proxy = self.mqtt_proxy

    @mock.patch("worker_bunch.utils.time_utils.TimeUtils.now")
    def test_round_trip(self, mocked_now):
        now = datetime(2022, 3, 19, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600)))
        mocked_now.return_value = now

        self.worker.subscribe_notifications(self.dispatcher)

        self.worker._work([Notification.create_cron("does-it-matter?")])

        expected_payload = self.manager.get_astral_times(now)
        expected_payload["status"] = "ok"

        self.mqtt_proxy.queue.assert_called_once_with(
            topic=AstralTimesPublisherConfKey.MQTT_TOPIC_OUT,
            payload=expected_payload,
            retain=True
        )



import copy
import unittest

from paho.mqtt.client import MQTTMessage

from worker_bunch.notification import Notification, NotificationType, NotificationBucket


class TestNotification(unittest.TestCase):

    def test_key(self):
        n1 = Notification(type=NotificationType.MQTT_MESSAGE, topic="1", payload="1")
        n21 = Notification(type=NotificationType.MQTT_MESSAGE, topic="2", payload="2.1")
        n22 = Notification(type=NotificationType.MQTT_MESSAGE, topic="2", payload="2.2")
        n3 = Notification(type=NotificationType.TIMER, topic="1", payload="1")

        self.assertEqual(n21, copy.deepcopy(n21))
        self.assertTrue(n21 == n22)
        self.assertTrue(n1 != n3)

    def test_create_from_mqtt(self):
        def create_msg(topic, payload):
            m = MQTTMessage(topic=topic)  # topic must be "bytes"
            m.payload = payload
            return m

        self.assertEqual(
            Notification.create_from_mqtt(create_msg(b"1", "1")),
            Notification(type=NotificationType.MQTT_MESSAGE, topic="1", payload="1")
        )

        self.assertEqual(
            Notification.create_from_mqtt(create_msg(b"1", b"1")),
            Notification(type=NotificationType.MQTT_MESSAGE, topic="1", payload="1")
        )


class TestNotificationBucket(unittest.TestCase):

    def test_store(self):
        n1 = Notification(type=NotificationType.MQTT_MESSAGE, topic="1", payload="1")
        n21 = Notification(type=NotificationType.MQTT_MESSAGE, topic="2", payload="2.1")
        n22 = Notification(type=NotificationType.MQTT_MESSAGE, topic="2", payload="2.2")

        bucket = NotificationBucket()

        def add(n: Notification):
            n = copy.deepcopy(n)
            bucket.add(n)

        add(n1)
        add(n21)
        add(n22)

        self.assertEqual(len(bucket), 2)

        store_list = bucket.get_list()
        store_list.sort(key=lambda n: n.topic)
        self.assertEqual(store_list[0], n1)

        # n21 get overwritten by a newer version
        self.assertEqual(store_list[1], n22)
        self.assertEqual(store_list[1].payload, n22.payload)

        self.assertTrue(bucket)

        bucket.clear()
        self.assertEqual(len(bucket), 0)
        self.assertFalse(bucket)



import asyncio
import logging
import signal
import threading
from asyncio import Task
from typing import List, Optional

from worker_bunch.dispatcher import Dispatcher
from worker_bunch.mqtt.mqtt_proxy import MqttProxy
from worker_bunch.notification import NotificationType
from worker_bunch.utils.time_utils import TimeUtils
from worker_bunch.worker.worker import Worker

_logger = logging.getLogger(__name__)


class Runner:

    TIME_LIMIT_MQTT_CONNECTION = 10  # seconds

    def __init__(self, dispatcher: Dispatcher, mqtt_proxy: MqttProxy, workers: List[Worker]):

        # init
        self._dispatcher = dispatcher
        self._mqtt_proxy = mqtt_proxy
        self._workers = workers

        self._loop = asyncio.get_event_loop()
        self._main_task = None  # type: Optional[Task]

        if threading.current_thread() is threading.main_thread():
            # integration tests may run the service in a thread...
            signal.signal(signal.SIGINT, self._shutdown_signaled)
            signal.signal(signal.SIGTERM, self._shutdown_signaled)

    def _shutdown_signaled(self, sig, _frame):
        _logger.info("shutdown signaled (%s)", sig)
        if self._main_task:
            self._main_task.cancel()

    def run(self):
        """endless loop"""

        # connect mqtt - part 1 - trigger
        self._mqtt_proxy.connect()

        self._main_task = self._loop.create_task(self._main_loop())

        try:
            self._loop.run_until_complete(self._main_task)
        except asyncio.CancelledError:
            _logger.debug("canceling...")

    def run_single(self):
        # connect mqtt - part 1 - trigger
        self._mqtt_proxy.connect()

        self._main_task = self._loop.create_task(self._main_single())

        self._loop.run_until_complete(self._main_task)

    async def _wait_for_mqtt_connection_timeout(self):
        timeout = self.TIME_LIMIT_MQTT_CONNECTION
        try:
            return await asyncio.wait_for(self._wait_for_mqtt_connection(), timeout)
        except asyncio.exceptions.TimeoutError:
            raise asyncio.exceptions.TimeoutError(f"couldn't connect to MQTT (within {timeout}s)!") from None

    async def _wait_for_mqtt_connection(self):
        """connect mqtt - part 2 - wait for connection and subscribe for topics"""
        while True:
            if self._mqtt_proxy.is_connected():
                for worker in self._workers:
                    worker.subscribe_notifications(self._dispatcher)

                topics = self._dispatcher.get_mqtt_topics()
                self._mqtt_proxy.subscribe(topics)
                break

            await asyncio.sleep(0.05)

    async def _main_loop(self):
        await self._wait_for_mqtt_connection_timeout()

        last_worker_check_time = TimeUtils.now()

        for worker in self._workers:
            worker.start()

        self._dispatcher.trigger_start_notification(self._workers)

        while True:
            self._mqtt_proxy.publish()

            self._mqtt_proxy.ensure_connection()
            messages = self._mqtt_proxy.get_messages()
            self._dispatcher.push_mqtt_messages(messages)

            self._dispatcher.trigger_timers()

            if (TimeUtils.now() - last_worker_check_time).total_seconds() > 20:
                last_worker_check_time = TimeUtils.now()
                dead_workers = [w.name for w in self._workers if not w.is_alive()]
                if dead_workers:
                    raise RuntimeError("Dead workers found: {}".format(", ".join(dead_workers)))

            await asyncio.sleep(0.05)

    async def _main_single(self):
        await self._wait_for_mqtt_connection_timeout()

        await asyncio.sleep(0.2)  # wait for messages to come in

        messages = self._mqtt_proxy.get_messages()
        self._dispatcher.push_mqtt_messages(messages)
        self._dispatcher.trigger_start_notification(self._workers, NotificationType.SINGLE_STARTED)

        await asyncio.sleep(0.2)  # wait for debounce pipelines to finish

        for worker in self._workers:
            worker.run_single()



import abc
import copy
import os
import unittest
from unittest import mock
from unittest.mock import MagicMock, call

import yaml
from psycopg.errors import DuplicateTable

from test.setup_test import SetupTest
from worker_bunch.database.database_config import DatabaseConfKey, MqttOutputType
from worker_bunch.mqtt.mqtt_client import MqttClient, MqttClientFactory
from worker_bunch.run_service import run_service
from worker_bunch.service_config import ConfigException


class MockMqttClientFactory:
    mock: MagicMock = None

    @classmethod
    def create(cls, _):
        if not cls.mock:
            cls.reset()
        return cls.mock

    @classmethod
    def reset(cls):
        cls.mock = MagicMock(MqttClient, autospec=True)


class BaseTestIntegration(unittest.TestCase):

    DO_LOGGING = False
    MAIN_WORKER = "mainWorker"

    def setUp(self):
        SetupTest.init_database()
        if self.DO_LOGGING:
            SetupTest.init_logging()

        MockMqttClientFactory.reset()

        database_config = copy.deepcopy(SetupTest.get_database_params())

        steps = self.create_steps()
        worker_replacements = self.create_worker_replacements()
        self._config_file = self.create_config_file(database_config, steps, worker_replacements)

        try:
            SetupTest.execute_commands(["CREATE TABLE dummy (id SERIAL PRIMARY KEY, text VARCHAR(128))"])
        except DuplicateTable:
            SetupTest.execute_commands(["DELETE FROM dummy"])

        SetupTest.execute_commands(["INSERT INTO dummy (text) VALUES ('text1')"])

        # just to make that all is working
        fetched = SetupTest.query_one("select COUNT(1) AS count FROM dummy")
        self.assertEqual(fetched["count"], 1)

        self._mocked_mqtt_client = MagicMock(MqttClient, autospec=True)

    def tearDown(self):
        SetupTest.close_database()

    @abc.abstractmethod
    def create_steps(self):
        raise NotImplementedError()

    def create_worker_replacements(self):
        return None

    def run_service_single_mode(self, config_file):
        run_service(config_file=config_file, log_file=None, log_level="debug", print_log_console=True, skip_log_times=True,
                    test_single=self.MAIN_WORKER)

    @classmethod
    def prepare_mocked_mqtt_client(cls, mocked_mqtt_client):
        mocked_mqtt_client.is_connected.return_value = True

    @classmethod
    def create_config_file(cls, database_config, steps, worker_replacements=None):
        worker_settings = {
            "connection_key": "main",
            "cron": "* * * * *",
            "steps": steps,
        }
        if worker_replacements:
            worker_settings["replacements"] = worker_replacements

        data = {
            "mqtt_broker": {"host": "mocked"},
            "database_connections": {"main": database_config},
            "worker_instances": {cls.MAIN_WORKER: "worker_bunch.database.database_worker.DatabaseWorker"},
            "worker_settings": {cls.MAIN_WORKER: worker_settings},
        }

        config_file = SetupTest.get_test_path("config_file.yaml")
        with open(config_file, 'w') as write_file:
            yaml.dump(data, write_file, default_flow_style=False)

        os.chmod(config_file, 0o600)

        return config_file


class TestIntegrationScalar(BaseTestIntegration):

    def create_steps(self):
        return [{
            DatabaseConfKey.MQTT_LAST_WILL: DatabaseConfKey.MQTT_LAST_WILL,
            DatabaseConfKey.MQTT_OUTPUT_TYPE: MqttOutputType.SCALAR,
            DatabaseConfKey.MQTT_TOPIC: DatabaseConfKey.MQTT_TOPIC,
            DatabaseConfKey.REPLACEMENTS: {"%XXX%": "dummy"},
            DatabaseConfKey.STATEMENT: "SELECT COUNT(1) FROM %XXX%",
        }]

    @mock.patch.object(MqttClientFactory, "create", MockMqttClientFactory.create)
    def test_full_integration_scalar(self):
        mocked_mqtt_client = MockMqttClientFactory.mock
        self.prepare_mocked_mqtt_client(mocked_mqtt_client)

        self.run_service_single_mode(self._config_file)

        mocked_mqtt_client.set_last_will.assert_called_once_with(
            topic=DatabaseConfKey.MQTT_TOPIC, last_will=DatabaseConfKey.MQTT_LAST_WILL, retain=False
        )

        calls = [
            call(topic=DatabaseConfKey.MQTT_TOPIC, payload="1", retain=False),
            call(topic=DatabaseConfKey.MQTT_TOPIC, payload=DatabaseConfKey.MQTT_LAST_WILL, retain=False),
        ]
        mocked_mqtt_client.publish.assert_has_calls(calls)


class TestIntegrationJson(BaseTestIntegration):
    """Test JSON + script file"""

    def setUp(self):
        statement = "SELECT '{%a%}' as a, '{%b%}' as b"
        script_file = SetupTest.get_test_path("integration_template_file.sql")
        with open(script_file, 'w') as f:
            f.write(statement)

        self.steps = [{
            DatabaseConfKey.MQTT_LAST_WILL: DatabaseConfKey.MQTT_LAST_WILL,
            DatabaseConfKey.MQTT_OUTPUT_TYPE: MqttOutputType.JSON,
            DatabaseConfKey.MQTT_TOPIC: DatabaseConfKey.MQTT_TOPIC,
            DatabaseConfKey.REPLACEMENTS: {"{%a%}": "a_step"},
            DatabaseConfKey.SCRIPT_FILE: script_file,
        }]

        self.worker_replacements = {"{%a%}": "a_worker", "{%b%}": "b_worker"}

        super().setUp()

    def create_steps(self):
        return self.steps

    def create_worker_replacements(self):
        return self.worker_replacements

    @mock.patch.object(MqttClientFactory, "create", MockMqttClientFactory.create)
    def test_full_integration_json(self):
        mocked_mqtt_client = MockMqttClientFactory.mock
        self.prepare_mocked_mqtt_client(mocked_mqtt_client)

        self.run_service_single_mode(self._config_file)

        mocked_mqtt_client.set_last_will.assert_called_once_with(
            topic=DatabaseConfKey.MQTT_TOPIC, last_will=DatabaseConfKey.MQTT_LAST_WILL, retain=False
        )

        calls = [
            call(topic=DatabaseConfKey.MQTT_TOPIC, payload='{"a": "a_step", "b": "b_worker"}', retain=False),
            call(topic=DatabaseConfKey.MQTT_TOPIC, payload=DatabaseConfKey.MQTT_LAST_WILL, retain=False),
        ]
        mocked_mqtt_client.publish.assert_has_calls(calls)


class TestIntegrationFileNotFound(BaseTestIntegration):
    """Test JSON + script file"""

    def setUp(self):
        self.script_file = SetupTest.get_test_path("integration_template_file.sql")
        if os.path.exists(self.script_file):
            os.remove(self.script_file)

        self.steps = [{
            DatabaseConfKey.MQTT_OUTPUT_TYPE: MqttOutputType.JSON,
            DatabaseConfKey.MQTT_TOPIC: DatabaseConfKey.MQTT_TOPIC,
            DatabaseConfKey.SCRIPT_FILE: self.script_file,
        }]

        super().setUp()

    def create_steps(self):
        return self.steps

    @mock.patch.object(MqttClientFactory, "create", MockMqttClientFactory.create)
    def test_full_integration_json(self):
        with self.assertRaises(ConfigException):
            self.run_service_single_mode(self._config_file)


class TestIntegrationExecute(BaseTestIntegration):

    ADD_ROWS = 6

    def create_steps(self):
        steps = []
        for i in range(self.ADD_ROWS):
            text = f"text-{i}"
            steps.append({
                DatabaseConfKey.MQTT_OUTPUT_TYPE: MqttOutputType.NONE,
                DatabaseConfKey.STATEMENT: f"INSERT INTO dummy (text) VALUES ('{text}')",
            })
        return steps

    @mock.patch.object(MqttClientFactory, "create", MockMqttClientFactory.create)
    def test_full_integration_execute(self):
        mocked_mqtt_client = MockMqttClientFactory.mock
        self.prepare_mocked_mqtt_client(mocked_mqtt_client)

        self.run_service_single_mode(self._config_file)

        fetched = SetupTest.query_one("select COUNT(1) AS count FROM dummy")
        self.assertEqual(fetched["count"], 7)



from typing import Dict

from worker_bunch.worker.worker import Worker
from worker_bunch.worker.worker_config import WorkerSettingsDeclaration


class WorkerFactory:

    PREDEFINED_WORKERS = {
        "AstralTimesPublisher": "worker_bunch.astral_times.astral_times_publisher.AstralTimesPublisher",
        "DatabaseWorker": "worker_bunch.database.database_worker.DatabaseWorker",
    }

    @classmethod
    def _check_class(cls, candidate):
        if not isinstance(candidate, Worker):
            if candidate:
                class_info = candidate.__class__.__module__ + '.' + candidate.__class__.__name__
            else:
                class_info = 'None'
            class_target = Worker.__module__ + '.' + Worker.__name__
            raise TypeError("{} is not of type {}!".format(class_info, class_target))

    @classmethod
    def resolve_import(cls, path: str) -> Worker.__class__:
        delimiter = path.rfind(".")
        class_name = path[delimiter + 1:len(path)]
        module_name = __import__(path[0:delimiter], globals(), locals(), [class_name])
        return getattr(module_name, class_name)

    @classmethod
    def create_worker(cls, worker_name: str, class_path: str):
        try:
            worker_class = cls.resolve_import(class_path)
            worker = worker_class(worker_name)
            cls._check_class(worker)
            return worker
        except Exception as ex:
            print('could not create worker "{}"! {}'.format(class_path, ex))
            raise

    @classmethod
    def create_workers(cls, config) -> Dict[str, Worker]:
        workers = {}
        for worker_name, class_path in config.items():

            predefined_class_paths = cls.PREDEFINED_WORKERS.get(class_path)
            if predefined_class_paths:
                class_path = predefined_class_paths

            if worker_name and class_path:
                workers[worker_name] = cls.create_worker(worker_name, class_path)
        return workers

    @classmethod
    def extract_workers_settings_declarations(cls, workers: Dict[str, Worker]) -> Dict[str, WorkerSettingsDeclaration]:
        settings_declarations = {}
        for worker_name, worker in workers.items():
            settings_declarations[worker_name] = WorkerSettingsDeclaration(
                settings_schema=worker.get_partial_settings_schema(),
                required=worker.make_partial_settings_required(),
            )
        return settings_declarations



import logging
import threading
from typing import List, Optional

import paho.mqtt.client as mqtt

from worker_bunch.mqtt.mqtt_config import MqttConfKey


_logger = logging.getLogger(__name__)


class MqttException(Exception):
    pass


class MqttClient:

    DEFAULT_KEEPALIVE = 60
    DEFAULT_PORT = 1883
    DEFAULT_PORT_SSL = 8883
    DEFAULT_PROTOCOL = 4  # 5==MQTTv5, default: 4==MQTTv311, 3==MQTTv31
    DEFAULT_QOS = 2
    DEFAULT_RETAIN = True

    TIME_WAIT_FOR_CONNECTION = 10  # seconds

    def __init__(self, config):

        self._host = None
        self._port = None
        self._keepalive = None

        self._client = None
        self._is_connected = False
        self._connection_error_info = None  # type: Optional[str]
        self._subscribed = False
        self._shutdown = False

        self._lock = threading.Lock()

        self._messages = []  # type: List[mqtt.MQTTMessage]

        self._host = config[MqttConfKey.HOST]
        self._port = config.get(MqttConfKey.PORT)
        self._keepalive = config.get(MqttConfKey.KEEPALIVE, self.DEFAULT_KEEPALIVE)

        self._default_qos = config.get(MqttConfKey.DEFAULT_QOS, self.DEFAULT_QOS)
        self._default_retain = config.get(MqttConfKey.DEFAULT_RETAIN, self.DEFAULT_RETAIN)

        self._debug_simulate_sending = config.get(MqttConfKey.DEBUG_SIMULATE_SENDING, False)
        self._debug_topic_prefix = config.get(MqttConfKey.DEBUG_TOPIC_PREFIX)

        protocol = config.get(MqttConfKey.PROTOCOL, self.DEFAULT_PROTOCOL)
        client_id = config.get(MqttConfKey.CLIENT_ID)
        ssl_ca_certs = config.get(MqttConfKey.SSL_CA_CERTS)
        ssl_certfile = config.get(MqttConfKey.SSL_CERTFILE)
        ssl_keyfile = config.get(MqttConfKey.SSL_KEYFILE)
        ssl_insecure = config.get(MqttConfKey.SSL_INSECURE, False)
        is_ssl = ssl_ca_certs or ssl_certfile or ssl_keyfile
        user_name = config.get(MqttConfKey.USER)
        user_pwd = config.get(MqttConfKey.PASSWORD)

        if not self._port:
            self._port = self.DEFAULT_PORT_SSL if is_ssl else self.DEFAULT_PORT

        self._client = mqtt.Client(client_id=client_id, protocol=protocol)

        if is_ssl:
            self._client.tls_set(ca_certs=ssl_ca_certs, certfile=ssl_certfile, keyfile=ssl_keyfile)
            if ssl_insecure:
                _logger.info("disabling SSL certificate verification")
                self._client.tls_insecure_set(True)

        if user_name or user_pwd:
            self._client.username_pw_set(user_name, user_pwd)

        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect
        self._client.on_message = self._on_message
        self._client.on_publish = self._on_publish

        self._client.reconnect_delay_set()

    def is_connected(self):
        with self._lock:
            return self._is_connected

    def connect(self):
        self._client.connect_async(self._host, port=self._port, keepalive=self._keepalive)
        self._client.loop_start()
        _logger.debug("%s is connecting...", self.__class__.__name__)

    def close(self):
        self._shutdown = True
        if self._client is not None:
            self._client.loop_stop()
            self._client.disconnect()
            self._client.loop_forever()  # will block until disconnect complete
            self._client = None
            _logger.debug("%s was closed.", self.__class__.__name__)

    def ensure_connection(self):
        """
        Check for rarely unexpected disconnects, but when happens, it's not clear how to heal. At least the loop has to be restarted.
        Best to restart the whole app. Recognise a stopped service in system log.
        """
        with self._lock:
            is_connected = self._is_connected
            connection_error_info = self._connection_error_info

        if connection_error_info:
            raise MqttException(connection_error_info)  # leads to exit => restarted by systemd
        if not is_connected:
            raise MqttException("MQTT is not connected!")

    def set_last_will(self, topic: str, last_will: str, retain: Optional[bool] = None, qos: Optional[int] = None):
        if self.is_connected():
            raise MqttException("MQTT last wills must be set before connecting!")

        retain = self._default_retain if retain is None else retain
        qos = self._default_qos if qos is None else qos

        self._client.will_set(
            topic=topic,
            payload=last_will,
            qos=qos,
            retain=retain
        )

        _logger.debug("set last will: topic='%s'; retain=%s; qos=%d; last_will='%s'", topic, retain, qos, last_will)

    def get_messages(self) -> List[mqtt.MQTTMessage]:
        with self._lock:
            messages = self._messages
            self._messages = []
            return messages

    def publish(self, topic: str, payload: str, retain: Optional[bool] = None, qos: Optional[int] = None):
        if self._shutdown:
            return

        retain = self._default_retain if retain is None else retain
        qos = self._default_qos if qos is None else qos

        if self._debug_topic_prefix:
            topic = self._debug_topic_prefix + topic

        if self._debug_simulate_sending:
            _logger.info("simulated sent: topic='%s'; retain=%s; qos=%d; payload='%s'", topic, retain, qos, payload)
            return mqtt.MQTTMessageInfo(0)

        result = self._client.publish(
            topic=topic,
            payload=payload,
            qos=qos,
            retain=retain
        )

        _logger.debug("sent: topic='%s'; retain=%s; qos=%d; payload='%s'", topic, retain, qos, payload)

        return result

    def subscribe(self, topics: List[str]):
        subs_qos = 1  # qos for subscriptions, not used, but necessary
        subscriptions = [(t, subs_qos) for t in topics]
        result, dummy = self._client.subscribe(subscriptions)
        if result != mqtt.MQTT_ERR_SUCCESS:
            error_info = "{} (#{})".format(mqtt.error_string(result), result)
            raise MqttException(f"could not subscribe to MQTT topics): {error_info}; topics: {topics}")

    def _on_connect(self, _mqtt_client, _userdata, _flags, rc):
        """MQTT callback is called when client connects to MQTT server."""
        class_name = self.__class__.__name__
        if rc == 0:
            with self._lock:
                self._is_connected = True
            _logger.debug("%s was connected.", class_name)
        else:
            connection_error_info = f"{class_name} connection failed (#{rc}: {mqtt.error_string(rc)})!"
            _logger.error(connection_error_info)
            with self._lock:
                self._is_connected = False
                self._connection_error_info = connection_error_info

    def _on_disconnect(self, _mqtt_client, _userdata, rc):
        """MQTT callback for when the client disconnects from the MQTT server."""
        class_name = self.__class__.__name__
        connection_error_info = None
        if rc != 0:
            connection_error_info = f"{class_name} connection was lost (#{rc}: {mqtt.error_string(rc)}) => abort => restart!"

        with self._lock:
            self._is_connected = False
            if connection_error_info and not self._connection_error_info:
                self._connection_error_info = connection_error_info

        if rc == 0:
            _logger.debug("%s was disconnected.", class_name)
        else:
            _logger.error("%s was unexpectedly disconnected: %s", class_name, connection_error_info or "???")

    def _on_message(self, _mqtt_client, _userdata, mqtt_message: mqtt.MQTTMessage):
        """MQTT callback when a message is received from MQTT server"""
        with self._lock:
            _logger.debug("on_message(%s): %s", mqtt_message.topic, mqtt_message.payload)
            self._messages.append(mqtt_message)

    def _on_publish(self, mqtt_client, userdata, mid):
        """MQTT callback is invoked when message was successfully sent to the MQTT server."""


class MqttClientFactory:
    """purpose: patch MqttClient within tests..."""

    @classmethod
    def create(cls, config):
        return MqttClient(config)



import unittest
from datetime import timezone, timedelta, datetime

from tzlocal import get_localzone

from worker_bunch.astral_times.astral_times_config import AstralTime, AstralTimesConfKey
from worker_bunch.astral_times.astral_times_manager import AstralTimesManager, AstralParsed


class TestAstralTimeManager(unittest.TestCase):

    DUMMY_CONFIG = {
        AstralTimesConfKey.LATITUDE: 51.051873,
        AstralTimesConfKey.LONGITUDE: 13.741522,
        AstralTimesConfKey.ELEVATION: 125,
    }

    def test_round_time_to_minute(self):
        tz = get_localzone()

        self.assertEqual(
            AstralTimesManager.round_time_to_minute(datetime(2022, 3, 19, 9, 55, 15, tzinfo=tz)),
            datetime(2022, 3, 19, 9, 55, 0, tzinfo=tz)
        )
        self.assertEqual(
            AstralTimesManager.round_time_to_minute(datetime(2022, 3, 19, 9, 55, 31, 12345, tzinfo=tz)),
            datetime(2022, 3, 19, 9, 56, 0, tzinfo=tz)
        )
        self.assertEqual(
            AstralTimesManager.round_time_to_minute(datetime(2022, 3, 19, 23, 59, 31, tzinfo=tz)),
            datetime(2022, 3, 20, 0, 0, 0, tzinfo=tz)
        )

    def test_astral_parse_type(self):
        a1 = AstralParsed(predefined=AstralTime.SUNRISE, is_dawn=True, is_dusk=True, depression=6)
        self.assertEqual(a1.is_valid(), False)

        # Check AstralParse to see if the prop initialisation is working properly
        a2 = AstralParsed()
        self.assertEqual(a2.predefined, None)
        self.assertEqual(a2.is_dawn, None)
        self.assertEqual(a2.is_dusk, None)
        self.assertEqual(a2.depression, None)
        self.assertEqual(a2.is_valid(), False)

        self.assertEqual(a1.predefined, AstralTime.SUNRISE)
        self.assertEqual(a1.is_dawn, True)
        self.assertEqual(a1.is_dusk, True)
        self.assertEqual(a1.depression, 6)

        self.assertEqual(AstralParsed(predefined=AstralTime.SUNRISE).is_valid(), True)
        self.assertEqual(AstralParsed(is_dawn=True, depression=6).is_valid(), True)
        self.assertEqual(AstralParsed(is_dusk=True, depression=6).is_valid(), True)
        self.assertEqual(AstralParsed(is_dawn=True, is_dusk=True, depression=6).is_valid(), False)
        self.assertEqual(AstralParsed(is_dusk=True).is_valid(), False)
        self.assertEqual(AstralParsed(is_dusk=True, depression=0).is_valid(), False)

    # noinspection SpellCheckingInspection
    def test_parse_astral_time_key(self):

        with self.assertRaises(ValueError):
            AstralTimesManager.parse_astral_time_key("dawn")
        with self.assertRaises(ValueError):
            AstralTimesManager.parse_astral_time_key("dawnnovalid")
        with self.assertRaises(ValueError):
            AstralTimesManager.parse_astral_time_key("dawn_22")

        self.assertEqual(
            AstralTimesManager.parse_astral_time_key("dawn_18"),
            AstralParsed(is_dawn=True, depression=18)
        )
        self.assertEqual(
            AstralTimesManager.parse_astral_time_key("dusk_2"),
            AstralParsed(is_dusk=True, depression=2)
        )
        self.assertEqual(
            AstralTimesManager.parse_astral_time_key("dusk_02"),
            AstralParsed(is_dusk=True, depression=2)
        )

        for astral_time in AstralTime:
            expected = AstralParsed(predefined=astral_time)
            got = AstralTimesManager.parse_astral_time_key(astral_time.value)
            self.assertEqual(got, expected)
            self.assertEqual(got.is_valid(), True)

        self.assertEqual(
            AstralTimesManager.parse_astral_time_key("  Sunrise "),
            AstralParsed(predefined=AstralTime.SUNRISE)
        )

    def test_get_location(self):
        m = AstralTimesManager(self.DUMMY_CONFIG)
        location = m.get_location()
        self.assertEqual(location.latitude, self.DUMMY_CONFIG[AstralTimesConfKey.LATITUDE])
        self.assertEqual(location.longitude, self.DUMMY_CONFIG[AstralTimesConfKey.LONGITUDE])
        self.assertEqual(location.elevation, self.DUMMY_CONFIG[AstralTimesConfKey.ELEVATION])

    def test_get_astral_time(self):
        m = AstralTimesManager(self.DUMMY_CONFIG)
        now = datetime(2022, 3, 19, 7, 55, 15, tzinfo=timezone(timedelta(seconds=3600)))

        sunset = m.get_astral_time("sunset", now)
        self.assertEqual(sunset, datetime(2022, 3, 19, 18, 18, tzinfo=timezone(timedelta(seconds=3600))))

        astral_times = m.get_astral_times(now)
        self.assertEqual(sunset, astral_times["sunset"])

        with self.assertRaises(ValueError):
            m.get_astral_time("does not exists", now)

    def test_get_astral_times(self):
        m = AstralTimesManager(self.DUMMY_CONFIG)
        now = datetime(2022, 3, 19, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600)))
        astral_times = m.get_astral_times(now)

        expected_times = {
            "timestamp": datetime(2022, 3, 19, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600))),
            "dawn_astro": datetime(2022, 3, 19, 4, 56, tzinfo=timezone(timedelta(seconds=3600))),
            "dawn_nautical": datetime(2022, 3, 19, 4, 16, tzinfo=timezone(timedelta(seconds=3600))),
            "dawn_civil": datetime(2022, 3, 19, 5, 35, tzinfo=timezone(timedelta(seconds=3600))),
            "sunrise": datetime(2022, 3, 19, 6, 8, tzinfo=timezone(timedelta(seconds=3600))),
            "noon": datetime(2022, 3, 19, 12, 13, tzinfo=timezone(timedelta(seconds=3600))),
            "sunset": datetime(2022, 3, 19, 18, 18, tzinfo=timezone(timedelta(seconds=3600))),
            "dusk_civil": datetime(2022, 3, 19, 18, 52, tzinfo=timezone(timedelta(seconds=3600))),
            "dusk_nautical": datetime(2022, 3, 19, 20, 11, tzinfo=timezone(timedelta(seconds=3600))),
            "dusk_astro": datetime(2022, 3, 19, 19, 31, tzinfo=timezone(timedelta(seconds=3600))),
            "midnight": datetime(2022, 3, 19, 0, 13, tzinfo=timezone(timedelta(seconds=3600)))
        }
        self.assertEqual(astral_times, expected_times)

    def test_get_astral_times_invalid(self):
        m = AstralTimesManager(self.DUMMY_CONFIG)
        now = datetime(2022, 6, 21, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600)))
        astral_times = m.get_astral_times(now)

        expected_times = {
            'dawn_astro': datetime(2022, 6, 21, 1, 46, tzinfo=timezone(timedelta(seconds=3600))),
            'dawn_civil': datetime(2022, 6, 21, 2, 59, tzinfo=timezone(timedelta(seconds=3600))),
            'dawn_nautical': None,
            'dusk_astro': datetime(2022, 6, 21, 22, 28, tzinfo=timezone(timedelta(seconds=3600))),
            'dusk_civil': datetime(2022, 6, 21, 21, 14, tzinfo=timezone(timedelta(seconds=3600))),
            'dusk_nautical': None,
            'midnight': datetime(2022, 6, 21, 0, 7, tzinfo=timezone(timedelta(seconds=3600))),
            'noon': datetime(2022, 6, 21, 12, 7, tzinfo=timezone(timedelta(seconds=3600))),
            'sunrise': datetime(2022, 6, 21, 3, 47, tzinfo=timezone(timedelta(seconds=3600))),
            'sunset': datetime(2022, 6, 21, 20, 26, tzinfo=timezone(timedelta(seconds=3600))),
            'timestamp': datetime(2022, 6, 21, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600)))
        }
        self.assertEqual(astral_times, expected_times)

    def test_get_astral_times_extented_invalid(self):
        m = AstralTimesManager(self.DUMMY_CONFIG)
        now = datetime(2022, 6, 21, 9, 55, 15, tzinfo=timezone(timedelta(seconds=7200)))

        keys = AstralTime.extended_values()

        astral_times = {}

        for key in keys:
            astral_times[key] = m.get_astral_time(key, now)

        self.assertEqual(astral_times, {
            'dawn_10': datetime(2022, 6, 21, 3, 14, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_11': datetime(2022, 6, 21, 3, 1, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_12': datetime(2022, 6, 21, 2, 46, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_13': datetime(2022, 6, 21, 2, 28, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_14': datetime(2022, 6, 21, 2, 6, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_15': datetime(2022, 6, 21, 1, 27, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_16': None,
            'dawn_17': None,
            'dawn_18': None,
            'dawn_2': datetime(2022, 6, 21, 4, 36, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_3': datetime(2022, 6, 21, 4, 27, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_4': datetime(2022, 6, 21, 4, 18, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_5': datetime(2022, 6, 21, 4, 9, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_6': datetime(2022, 6, 21, 3, 59, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_7': datetime(2022, 6, 21, 3, 49, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_8': datetime(2022, 6, 21, 3, 38, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_9': datetime(2022, 6, 21, 3, 27, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_astro': datetime(2022, 6, 21, 2, 46, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_civil': datetime(2022, 6, 21, 3, 59, tzinfo=timezone(timedelta(seconds=7200))),
            'dawn_nautical': None,
            'dusk_10': datetime(2022, 6, 21, 22, 59, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_11': datetime(2022, 6, 21, 23, 13, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_12': datetime(2022, 6, 21, 23, 28, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_13': datetime(2022, 6, 21, 23, 46, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_14': datetime(2022, 6, 22, 0, 8, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_15': datetime(2022, 6, 22, 0, 47, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_16': None,
            'dusk_17': None,
            'dusk_18': None,
            'dusk_2': datetime(2022, 6, 21, 21, 38, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_3': datetime(2022, 6, 21, 21, 46, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_4': datetime(2022, 6, 21, 21, 55, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_5': datetime(2022, 6, 21, 22, 5, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_6': datetime(2022, 6, 21, 22, 14, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_7': datetime(2022, 6, 21, 22, 25, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_8': datetime(2022, 6, 21, 22, 36, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_9': datetime(2022, 6, 21, 22, 47, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_astro': datetime(2022, 6, 21, 23, 28, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_civil': datetime(2022, 6, 21, 22, 14, tzinfo=timezone(timedelta(seconds=7200))),
            'dusk_nautical': None,
            'midnight': datetime(2022, 6, 21, 1, 7, tzinfo=timezone(timedelta(seconds=7200))),
            'noon': datetime(2022, 6, 21, 13, 7, tzinfo=timezone(timedelta(seconds=7200))),
            'sunrise': datetime(2022, 6, 21, 4, 47, tzinfo=timezone(timedelta(seconds=7200))),
            'sunset': datetime(2022, 6, 21, 21, 26, tzinfo=timezone(timedelta(seconds=7200))),
        })

    def test_hits(self):
        m = AstralTimesManager(self.DUMMY_CONFIG)
        t_no_hit = datetime(2022, 3, 19, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600)))
        t_sunrise = datetime(2022, 3, 19, 6, 8, tzinfo=timezone(timedelta(seconds=3600)))

        self.assertEqual(len(m._cached_values), 0)
        hit = m.hits(" Sunrise ", t_no_hit)
        self.assertEqual(hit, False)
        self.assertEqual(len(m._cached_values), 1)

        hit = m.hits(" Sunrise ", t_sunrise)
        self.assertEqual(hit, True)
        self.assertEqual(len(m._cached_values), 1)

        # noinspection SpellCheckingInspection
        hit = m.hits(" sunrisE ", t_sunrise)
        self.assertEqual(hit, True)
        self.assertEqual(len(m._cached_values), 2)

        # another hour => cache ist reset
        hit = m.hits("       Sunrise ", t_no_hit)
        self.assertEqual(hit, False)
        self.assertEqual(len(m._cached_values), 1)  # cache was reset



import unittest
import uuid
from datetime import timedelta, timezone, datetime
from unittest.mock import MagicMock, call

from worker_bunch.mqtt.mqtt_client import MqttClient
from worker_bunch.mqtt.mqtt_proxy import MqttProxy


class TestMqttProxy(unittest.TestCase):

    def test_roundtrip_strings(self):
        client = MagicMock(MqttClient, autospec=True)
        proxy = MqttProxy(client)

        calls = []

        for i in range(1, 20):
            topic = f"{i}-test123"
            payload = f"{i}-{uuid.uuid4()}"
            retain = True if i % 2 == 0 else False

            proxy.queue(topic=topic, payload=payload, retain=retain)
            calls.append(call(topic=topic, payload=payload, retain=retain))

        proxy.publish()
        client.publish.assert_has_calls(calls)

    def test_roundtrip_dict(self):
        client = MagicMock(MqttClient, autospec=True)
        proxy = MqttProxy(client)

        topic = "test123"
        payload_to_convert = {
            "timestamp": datetime(2022, 3, 19, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600))),
            "boolean": True,
            "integer": 456,
            "float": 1.234,
            "text": "text123",
        }
        payload = '{"boolean": true, "float": 1.234, "integer": 456, "text": "text123", "timestamp": "2022-03-19T09:55:15+01:00"}'
        retain = True

        proxy.queue(topic=topic, payload=payload_to_convert, retain=retain)
        proxy.publish()

        client.publish.assert_called_once_with(topic=topic, payload=payload, retain=retain)



import copy
import locale
import logging
import os
import pathlib
import uuid
from typing import Dict, Optional

import yaml
from jsonschema import validate

from worker_bunch.service_config import MainConfKey, CONFIG_JSONSCHEMA, ServiceConfKey, ConfigException
from worker_bunch.worker.worker_config import WorkerSettingsDeclaration

_logger = logging.getLogger(__name__)


class ServiceConfigurator:

    def __init__(self):
        self._config_data = {}
        self._config_file: Optional[str] = None

    def read_config_file(self, config_file, skip_file_access_check=False):
        self._config_file = os.path.abspath(config_file)

        if not skip_file_access_check:
            self.check_config_file_access(self._config_file)

        with open(self._config_file, 'r') as stream:
            file_data = yaml.unsafe_load(stream)

        # first validation without worker config
        validate(file_data, CONFIG_JSONSCHEMA)

        self._config_data = file_data

        # ensure service section, which will be made available to workers
        service_settings = self._config_data.get(MainConfKey.SERVICE)
        if service_settings is None:
            service_settings = {}
            self._config_data[MainConfKey.SERVICE] = service_settings

    def init_locale(self):
        locale_name = self.get_service_config().get(ServiceConfKey.LOCALE)

        try:
            if locale_name:
                locale_was_set = locale.setlocale(locale.LC_ALL, locale_name.strip())
                _logger.debug("set locale: '%s'", locale_was_set)
        except locale.Error as ex:
            raise ConfigException(f"set locale failed (use e.g. 'de_DE.UTF8', not '{locale_name}'): {ex}")

    def init_data_dir(self):
        service_settings = self.get_service_config()
        data_dir = service_settings.get(ServiceConfKey.DATA_DIR)

        if not data_dir:
            _logger.info("no data dir configured. workers may not be able to write to disc.")
            return

        if not os.path.isabs(data_dir):
            dirname = os.path.dirname(self._config_file)
            data_dir = os.path.realpath(os.path.join(dirname, data_dir))
            service_settings[ServiceConfKey.DATA_DIR] = data_dir

        if not os.path.isdir(data_dir):
            try:
                os.makedirs(data_dir, exist_ok=True)
            except PermissionError as ex:
                raise ConfigException(f"Provide a proper, writable data directory! '{data_dir}': {ex}")

        # check write permission
        for i in range(1, 3):
            touch_file = os.path.join(data_dir, f"touch-test.{str(uuid.uuid4())}")
            if not os.path.exists(touch_file):
                try:
                    pathlib.Path(touch_file).touch()
                    os.remove(touch_file)
                    break
                except Exception as ex:
                    raise ConfigException(f"Data dir write check failed! {ex}")
        # ... _logger.info("could not check data dir write permission...?!")

    @classmethod
    def create_extended_json_schema(cls, declarations: Dict[str, WorkerSettingsDeclaration]) -> Dict[str, any]:
        schema = copy.deepcopy(CONFIG_JSONSCHEMA)

        worker_settings_schema = {"type": "object"}
        worker_settings_schema["additionalProperties"] = False

        settings_entries = {}
        worker_settings_schema["properties"] = settings_entries
        settings_required = []
        worker_settings_schema["required"] = settings_required

        for worker_name, extra_config in declarations.items():
            if extra_config.settings_schema:
                settings_entries[worker_name] = extra_config.settings_schema
                if extra_config.required:
                    settings_required.append(worker_name)

        schema["properties"][MainConfKey.WORKER_SETTINGS] = worker_settings_schema
        return schema

    def revalidate_worker_settings(self, declarations: Dict[str, WorkerSettingsDeclaration]):
        extended_schema = self.create_extended_json_schema(declarations)
        validate(self._config_data, extended_schema)

    def get_worker_instances_config(self):
        return self._config_data.get(MainConfKey.WORKER_INSTANCES, {})

    def get_worker_settings(self):
        return self._config_data.get(MainConfKey.WORKER_SETTINGS, {})

    def get_logging_config(self):
        return self._config_data.get(MainConfKey.LOGGING, {})

    def get_mqtt_config(self):
        return self._config_data.get(MainConfKey.MQTT_BROKER, {})

    def get_astral_config(self):
        return self._config_data.get(MainConfKey.ASTRAL_TIMES, {})

    def get_database_config(self):
        return self._config_data.get(MainConfKey.DATABASE_CONNECTIONS, {})

    def get_service_config(self):
        return self._config_data[MainConfKey.SERVICE]

    def get_data_dir(self) -> str:
        service_settings = self.get_service_config()
        return service_settings.get(ServiceConfKey.DATA_DIR)

    @classmethod
    def check_config_file_access(cls, config_file):
        if not os.path.isfile(config_file):
            raise FileNotFoundError('config file ({}) does not exist!'.format(config_file))

        permissions = oct(os.stat(config_file).st_mode & 0o777)[2:]
        if permissions != "600":
            extra = "change via 'chmod'. this config file may contain sensitive information."
            raise PermissionError(f"wrong config file permissions ({config_file}: expected 600, got {permissions})! {extra}")

    @classmethod
    def generate_config_file_json_schema(cls, workers_extra_configs: Optional[Dict[str, WorkerSettingsDeclaration]]):
        if workers_extra_configs:
            schema = cls.create_extended_json_schema(workers_extra_configs)
        else:
            schema = CONFIG_JSONSCHEMA
        return schema



import abc
import datetime
import logging
from logging import Logger
from typing import Optional

import psycopg
from tzlocal import get_localzone

from worker_bunch.database.database_config import DatabaseConfKey
from worker_bunch.service_logging import ServiceLogging
from worker_bunch.utils.time_utils import TimeUtils


class DatabaseException(Exception):
    pass


class DatabaseConnector(abc.ABC):

    def __init__(self, config, context_name: str, connection_key: str):

        self._context_name = context_name
        self._connection_key = connection_key
        self.__logger = None  # type: Optional[Logger]

        self._connection = None
        self._auto_commit = config.get(DatabaseConfKey.AUTO_COMMIT, False)
        self._last_connect_time = None  # type: Optional[datetime.datetime]

        # configuration
        self._connect_data = {
            "host": config[DatabaseConfKey.HOST],
            "port": config[DatabaseConfKey.PORT],
            "user": config[DatabaseConfKey.USER],
            "password": config.get(DatabaseConfKey.PASSWORD),
            "dbname": config[DatabaseConfKey.DATABASE],
        }

        self._timezone = config.get(DatabaseConfKey.TIMEZONE)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        self.close()

    @property
    def _logger(self):
        if self.__logger is None:
            log_name = f"{self._context_name}-{self._connection_key}"
            log_name = ServiceLogging.get_log_name(self, log_name)
            self.__logger = logging.getLogger(log_name)
        return self.__logger

    @property
    def is_connected(self):
        return bool(self._connection)

    @property
    def connection(self):
        return self._connection

    def connect(self):
        if self._connection:
            self._connection.close()

        try:
            self._connection = psycopg.connect(**self._connect_data, autocommit=self._auto_commit)

            with self._connection.cursor() as cursor:
                time_zone = self._timezone if self._timezone else self.get_default_time_zone_name()
                stmt = "set timezone='{}'".format(time_zone)
                try:
                    cursor.execute(stmt)
                except Exception:
                    self._logger.error("setting timezone failed (%s)!", stmt)
                    raise

            self._last_connect_time = TimeUtils.now()

        except psycopg.OperationalError as ex:
            raise DatabaseException(str(ex)) from ex

    def close(self):
        try:
            if self._connection:
                self._connection.close()
        except Exception as ex:
            self._logger.exception(ex)
        finally:
            self._connection = None

    @classmethod
    def get_default_time_zone_name(cls):
        local_timezone = get_localzone()
        if not local_timezone:
            local_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
        return str(local_timezone)




class MqttConfKey:
    CLIENT_ID = "client_id"
    HOST = "host"
    PORT = "port"
    PASSWORD = "password"
    USER = "user"
    KEEPALIVE = "keepalive"
    PROTOCOL = "protocol"
    DEFAULT_QOS = "default_qos"
    DEFAULT_RETAIN = "default_retain"

    DEBUG_SIMULATE_SENDING = "debug_simulate_sending"
    DEBUG_TOPIC_PREFIX = "debug_topic_prefix"

    SSL_CA_CERTS = "ssl_ca_certs"
    SSL_CERTFILE = "ssl_certfile"
    SSL_INSECURE = "ssl_insecure"
    SSL_KEYFILE = "ssl_keyfile"


MQTT_JSONSCHEMA = {
    "type": "object",
    "properties": {
        MqttConfKey.CLIENT_ID: {"type": "string", "minLength": 1},
        MqttConfKey.DEFAULT_QOS: {"type": "integer", "enum": [0, 1, 2]},
        MqttConfKey.DEFAULT_RETAIN: {
            "type": "boolean",
            "description": "Default: True. May be overwritten in your worker."},
        MqttConfKey.HOST: {"type": "string", "minLength": 1},
        MqttConfKey.KEEPALIVE: {
            "type": "integer",
            "minimum": 1,
            "description": "Maximum period in seconds between communications with the broker"
        },
        MqttConfKey.PASSWORD: {"type": "string"},
        MqttConfKey.PORT: {"type": "integer"},
        MqttConfKey.PROTOCOL: {"type": "integer", "enum": [3, 4, 5]},
        MqttConfKey.SSL_CA_CERTS: {"type": "string", "minLength": 1},
        MqttConfKey.SSL_CERTFILE: {"type": "string", "minLength": 1},
        MqttConfKey.SSL_INSECURE: {"type": "boolean"},
        MqttConfKey.SSL_KEYFILE: {"type": "string", "minLength": 1},
        MqttConfKey.USER: {"type": "string", "minLength": 1},

        MqttConfKey.DEBUG_SIMULATE_SENDING: {"type": "boolean", "description": "it True, no MQTT message is sent out!"},
        MqttConfKey.DEBUG_TOPIC_PREFIX: {
            "type": "string",
            "description": "Put before the intented topic!"
        },
    },
    "additionalProperties": False,
    "required": [MqttConfKey.HOST],
}



import unittest
from datetime import datetime, timezone, timedelta

from worker_bunch.utils.json_utils import JsonUtils


class TestJsonUtils(unittest.TestCase):

    def test_dumps(self):

        original_data = {
            "timestamp": datetime(2022, 3, 19, 9, 55, 15, tzinfo=timezone(timedelta(seconds=3600))),
            "boolean": True,
            "integer": 456,
            "float": 1.234,
            "text": "text123",
        }
        expected_data = '{"boolean": true, "float": 1.234, "integer": 456, "text": "text123", "timestamp": "2022-03-19T09:55:15+01:00"}'

        result_data = JsonUtils.dumps(original_data)

        self.assertEqual(result_data, expected_data)



from typing import Dict, Optional, List

from worker_bunch.astral_times.astral_times_manager import AstralTimesManager
from worker_bunch.dispatcher import Dispatcher
from worker_bunch.notification import Notification, NotificationType
from worker_bunch.worker.worker import Worker, WorkerSetup


class AstralTimesPublisherConfKey:
    MQTT_TOPIC_OUT = "mqtt_topic_out"
    MQTT_LAST_WILL = "mqtt_last_will"
    MQTT_RETAIN = "mqtt_retain"


ASTRAL_TIMES_PUBLISHER_JSONSCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "properties": {
        AstralTimesPublisherConfKey.MQTT_LAST_WILL: {"type": "string", "minLength": 1, "description": "MQTT last will message"},
        AstralTimesPublisherConfKey.MQTT_RETAIN: {"type": "boolean", "description": "Make MQTT message persistent."},
        AstralTimesPublisherConfKey.MQTT_TOPIC_OUT: {"type": "string", "minLength": 1, "description": "MQTT topic to write to"},
    },
    "required": [AstralTimesPublisherConfKey.MQTT_TOPIC_OUT]
}


class AstralTimesPublisher(Worker):

    def __init__(self, name: str):
        super().__init__(name)

        self._mqtt_last_will = ""
        self._mqtt_retain = False
        self._mqtt_topic_out = []
        self._mqtt_topics_in = ""

        self._astral_time_manager: Optional[AstralTimesManager] = None

    def setup(self, props):
        super().setup(props)

        self._astral_time_manager = props[WorkerSetup.ASTRAL_TIME_MANAGER]

        self._mqtt_last_will = self._worker_settings.get(AstralTimesPublisherConfKey.MQTT_LAST_WILL)
        self._mqtt_retain = self._worker_settings.get(AstralTimesPublisherConfKey.MQTT_RETAIN, False)
        self._mqtt_topic_out = self._worker_settings[AstralTimesPublisherConfKey.MQTT_TOPIC_OUT]

    def get_partial_settings_schema(self) -> Dict[str, any]:
        return ASTRAL_TIMES_PUBLISHER_JSONSCHEMA

    def make_partial_settings_required(self) -> bool:
        return True

    def set_last_will(self):
        """Set the last will at mqtt client. Explicitly triggered before the mqtt client connects."""
        if self._mqtt_topic_out and self._mqtt_last_will:
            self._mqtt_proxy.set_last_will(self._mqtt_topic_out, self._mqtt_last_will)

    def subscribe_notifications(self, dispatcher: Dispatcher):
        dispatcher.subscribe_cron(self, "0 * * * *", "cron-every-hour")

    def _work(self, notifications: List[Notification]):
        found_message = next((n for n in notifications if n.type in [NotificationType.CRON, NotificationType.JUST_STARTED]), None)

        if found_message:
            if self._astral_time_manager:
                astral_times = self._astral_time_manager.get_astral_times()
                astral_times["status"] = "ok"
                self._mqtt_proxy.queue(topic=self._mqtt_topic_out, payload=astral_times, retain=self._mqtt_retain)
            else:
                self._final_work()

    def _final_work(self):
        """sends the last will"""
        if self._mqtt_topic_out and self._mqtt_last_will:
            self._mqtt_proxy.queue(self._mqtt_topic_out, self._mqtt_last_will, retain=self._mqtt_retain)



import threading
from collections import namedtuple
from typing import Dict, List, Optional, Union

from paho.mqtt.client import MQTTMessage

from worker_bunch.mqtt.mqtt_client import MqttClient
from worker_bunch.service_config import ConfigException
from worker_bunch.utils.json_utils import JsonUtils

ProxyMessage = namedtuple("ProxyMessage", ["topic", "payload", "retain"])


class MqttProxy:

    def __init__(self, mqtt_client: Optional[MqttClient]):

        self._mqtt_client = mqtt_client
        self._lock = threading.Lock()

        self._messages: List[ProxyMessage] = []

    def close(self):
        self.publish()
        self._mqtt_client = None

    def connect(self):
        with self._lock:
            if self._mqtt_client:
                self._mqtt_client.connect()

    def is_connected(self):
        with self._lock:
            if self._mqtt_client:
                return self._mqtt_client.is_connected()
            else:
                return True  # hide missing mqtt client

    def ensure_connection(self):
        if self._mqtt_client:
            self._mqtt_client.ensure_connection()

    def set_last_will(self, topic: str, last_will: Union[str, Dict], retain: Optional[bool] = None):
        if not self._mqtt_client:
            raise ConfigException("no mqtt client configured!")

        if isinstance(last_will, dict):
            last_will = JsonUtils.dumps(last_will)

        with self._lock:
            if self._mqtt_client and topic and last_will:
                self._mqtt_client.set_last_will(topic=topic, last_will=last_will, retain=retain)

    def subscribe(self, topics: List[str]):
        if topics:
            if not self._mqtt_client:
                raise ConfigException("no mqtt client configured!")

            with self._lock:
                self._mqtt_client.subscribe(topics)

    def get_messages(self) -> List[MQTTMessage]:
        with self._lock:
            if self._mqtt_client:
                return self._mqtt_client.get_messages()
            else:
                return []

    def queue(self, topic: str, payload: Union[str, Dict], retain: Optional[bool] = None):
        if not self._mqtt_client:
            raise ConfigException("no mqtt client configured!")

        if isinstance(payload, dict):
            payload = JsonUtils.dumps(payload)

        with self._lock:
            self._messages.append(ProxyMessage(topic=topic, payload=payload, retain=retain))

    def publish(self):
        if self._mqtt_client:
            with self._lock:
                if self._mqtt_client:
                    messages = self._messages
                    self._messages = []
                    for m in messages:
                        self._mqtt_client.publish(topic=m.topic, payload=m.payload, retain=m.retain)



import os
import unittest

import yaml

from test.setup_test import SetupTest
from worker_bunch.database.database_config import DatabaseConfKey, MqttOutputType
from worker_bunch.run_service import generate_json_schema_info


class TestGenerateJsonSchemaInfo(unittest.TestCase):

    MAIN_WORKER = "mainWorker"

    @classmethod
    def create_config_file(cls, steps):
        data = {
            "mqtt_broker": {"host": "mocked"},
            "database_connections": {"main-db": {
                "host": "host",
                "port": 5435,
                "database": "database"
            }},
            "worker_instances": {cls.MAIN_WORKER: "worker_bunch.database.database_worker.DatabaseWorker"},
            "worker_settings": {cls.MAIN_WORKER: {
                "connection_key": "main",
                "cron": "* * * * *",
                "steps": steps,
            }},
        }

        config_file = SetupTest.get_test_path("config_file.yaml")
        with open(config_file, 'w') as write_file:
            yaml.dump(data, write_file, default_flow_style=False)

        os.chmod(config_file, 0o600)

        return config_file

    def assert_step_properties(self, output: str):
        self.assertIn(DatabaseConfKey.MQTT_LAST_WILL, output)
        self.assertIn(DatabaseConfKey.MQTT_OUTPUT_TYPE, output)
        self.assertIn(DatabaseConfKey.MQTT_TOPIC, output)
        self.assertIn(DatabaseConfKey.REPLACEMENTS, output)
        self.assertIn(DatabaseConfKey.STATEMENT, output)
        self.assertIn(DatabaseConfKey.SCRIPT_FILE, output)

    def test_complex(self):
        output = generate_json_schema_info(None)
        lines_without_config_file = len(output.split("\n"))
        self.assertGreater(lines_without_config_file, 100)

        steps = [{
            DatabaseConfKey.MQTT_LAST_WILL: DatabaseConfKey.MQTT_LAST_WILL,
            DatabaseConfKey.MQTT_OUTPUT_TYPE: MqttOutputType.SCALAR,
            DatabaseConfKey.MQTT_TOPIC: DatabaseConfKey.MQTT_TOPIC,
            DatabaseConfKey.REPLACEMENTS: {"%XXX%": "dummy"},
            DatabaseConfKey.STATEMENT: "SELECT COUNT(1) FROM %XXX%",
        }]
        config_file = self.create_config_file(steps)
        output = generate_json_schema_info(config_file)
        lines_with_config_file = len(output.split("\n"))
        self.assertGreater(lines_with_config_file, lines_without_config_file)
        self.assert_step_properties(output)

        steps = []
        config_file = self.create_config_file(steps)
        output = generate_json_schema_info(config_file)
        lines_with_invalid_config_file = len(output.split("\n"))
        self.assertGreaterEqual(lines_with_invalid_config_file, lines_with_config_file)

        self.assert_step_properties(output)





class DatabaseConfKey:

    # database connection
    HOST = "host"
    USER = "user"
    PORT = "port"
    PASSWORD = "password"
    DATABASE = "database"
    TIMEZONE = "timezone"
    AUTO_COMMIT = "auto_commit"

    # database worker
    CONNECTION_KEY = "connection_key"
    CRON = "cron"
    STEPS = "steps"

    # database step
    MQTT_LAST_WILL = "mqtt_last_will"
    MQTT_OUTPUT_TYPE = "mqtt_output_type"
    MQTT_RETAIN = "mqtt_retain"
    MQTT_TOPIC = "mqtt_topic"
    REPLACEMENTS = "replacements"
    SCRIPT_FILE = "script_file"
    STATEMENT = "statement"


DATABASE_CONNECTION_JSONSCHEMA = {
    "additionalProperties": False,
    "required": [DatabaseConfKey.HOST, DatabaseConfKey.PORT, DatabaseConfKey.DATABASE],
    "type": "object",
    "properties": {
        DatabaseConfKey.HOST: {"type": "string", "minLength": 1, "description": "Database host"},
        DatabaseConfKey.PORT: {"type": "integer", "minimum": 1, "description": "Database port"},
        DatabaseConfKey.USER: {"type": "string", "minLength": 1, "description": "Database user"},
        DatabaseConfKey.PASSWORD: {"type": "string", "minLength": 1, "description": "Database password"},
        DatabaseConfKey.DATABASE: {"type": "string", "minLength": 1, "description": "Database name"},
        DatabaseConfKey.TIMEZONE: {"type": "string", "minLength": 1, "description": "Predefined session timezone"},
        DatabaseConfKey.AUTO_COMMIT: {"type": "boolean"},
    },
}


DATABASE_CONNECTIONS_JSONSCHEMA = {
    "type": "object",
    "additionalProperties": DATABASE_CONNECTION_JSONSCHEMA,
    "description": "Dictionary of <database-connection-name>:<database-connection-properties>"
}


class MqttOutputType:
    JSON = "json"
    NONE = "none"
    SCALAR = "scalar"


MQTT_OUTPUT_TYPES = [MqttOutputType.JSON, MqttOutputType.NONE, MqttOutputType.SCALAR]


DATABASE_STEP_JSONSCHEMA = {
    "additionalProperties": False,
    "oneOf": [{"required": [DatabaseConfKey.STATEMENT]}, {"required": [DatabaseConfKey.SCRIPT_FILE]}],
    "type": "object",
    "properties": {
        DatabaseConfKey.MQTT_LAST_WILL: {"type": "string", "minLength": 1, "description": "MQTT last will"},
        DatabaseConfKey.MQTT_OUTPUT_TYPE: {
            "type": "string",
            "enum": MQTT_OUTPUT_TYPES,
            "description": "'scalar': publishes a single stringified value; 'json': publishes JSON (first row); 'none': publishes nothing",
        },
        DatabaseConfKey.MQTT_RETAIN: {"type": "boolean"},
        DatabaseConfKey.MQTT_TOPIC: {"type": "string", "minLength": 1, "description": "MQTT target topic"},
        DatabaseConfKey.SCRIPT_FILE: {"type": "string", "minLength": 1, "description": "SQL script file (statement has priority)"},
        DatabaseConfKey.STATEMENT: {"type": "string", "minLength": 1, "description": "SQL statement"},
        DatabaseConfKey.REPLACEMENTS: {
            "additionalProperties": {"type": "string"},
            "description": "Key/value pairs: replace all occurrences of the key with the value within statement.",
            "type": "object",
        },
    },
}


DATABASE_WORKER_JSONSCHEMA = {
    "additionalProperties": False,
    "required": [DatabaseConfKey.CONNECTION_KEY, DatabaseConfKey.CRON, DatabaseConfKey.STEPS],
    "type": "object",
    "properties": {
        DatabaseConfKey.CONNECTION_KEY: {"type": "string", "minLength": 1, "description": "Database connection key"},
        # TODO https://stackoverflow.com/questions/14203122/create-a-regular-expression-for-cron-statement
        DatabaseConfKey.CRON: {"type": "string", "minLength": 9, "description": "CRON syntax"},
        DatabaseConfKey.REPLACEMENTS: {
            "additionalProperties": {"type": "string"},
            "description": "Key/value pairs: Worker wide configuration. May be overwritten by steps. "
                           "Replace all occurrences of the key with the value within statements.",
            "type": "object",
        },
        DatabaseConfKey.STEPS: {
            "type": "array",
            "items": DATABASE_STEP_JSONSCHEMA,
            "description": "List of MQTT topics to read from",
            "minItems": 1,
        }
    },
}



import logging
import os
import pathlib
import sys
from typing import List, Optional

import psycopg
# noinspection PyPackageRequirements
import testing.postgresql
from psycopg.rows import dict_row


class SetupTestException(Exception):
    pass


_logger = logging.getLogger(__name__)


# noinspection SpellCheckingInspection
class SetupTest:

    TEST_DIR = "__test__"
    DATABASE_DIR = os.path.join(TEST_DIR, "database")

    _logging_inited = False
    _postgresql: Optional[testing.postgresql.Postgresql] = None

    @classmethod
    def init_logging(cls):
        if not cls._logging_inited:
            cls._logging_inited = True

            logging.basicConfig(
                format='[%(levelname)8s] %(name)s: %(message)s',
                level=logging.DEBUG,
                handlers=[logging.StreamHandler(sys.stdout)]
            )

            logging.getLogger("asyncio").setLevel(logging.WARNING)

    @classmethod
    def get_project_dir(cls) -> str:
        file_path = os.path.dirname(__file__)
        out = os.path.dirname(file_path)  # go up one time
        return out

    @classmethod
    def get_test_dir(cls) -> str:
        project_dir = cls.get_project_dir()
        out = os.path.join(project_dir, cls.TEST_DIR)
        return out

    @classmethod
    def get_test_path(cls, relative_path) -> str:
        return os.path.join(cls.get_test_dir(), relative_path)

    @classmethod
    def get_database_dir(cls) -> str:
        project_dir = cls.get_project_dir()
        out = os.path.join(project_dir, cls.DATABASE_DIR)
        return out

    @classmethod
    def ensure_test_dir(cls) -> str:
        return cls.ensure_dir(cls.get_test_dir())

    @classmethod
    def ensure_clean_test_dir(cls) -> str:
        return cls.ensure_clean_dir(cls.get_test_dir())

    @classmethod
    def ensure_database_dir(cls) -> str:
        return cls.ensure_dir(cls.get_database_dir())

    @classmethod
    def ensure_clean_database_dir(cls) -> str:
        return cls.ensure_clean_dir(cls.get_database_dir())

    @classmethod
    def ensure_dir(cls, dirpath) -> str:
        exists = os.path.exists(dirpath)

        if exists and not os.path.isdir(dirpath):
            raise NotADirectoryError(dirpath)
        if not exists:
            os.makedirs(dirpath)

        return dirpath

    @classmethod
    def ensure_clean_dir(cls, dirpath) -> str:
        if not os.path.exists(dirpath):
            cls.ensure_dir(dirpath)
        else:
            cls.clean_dir_recursively(dirpath)

        return dirpath

    @classmethod
    def clean_dir_recursively(cls, path_in):
        dir_segments = pathlib.Path(path_in)
        if not dir_segments.is_dir():
            return
        for item in dir_segments.iterdir():
            if item.is_dir():
                cls.clean_dir_recursively(item)
                os.rmdir(item)
            else:
                item.unlink()

    @classmethod
    def init_database(cls, recreate=False):
        database_dir = cls.get_database_dir()

        if recreate:
            cls.close_database(shutdown=True)

        if not cls._postgresql:
            cls.ensure_clean_dir(database_dir)
            cls._postgresql = testing.postgresql.Postgresql(base_dir=database_dir)

    @classmethod
    def close_database(cls, shutdown=False):
        if cls._postgresql:
            if shutdown:
                cls._postgresql.stop()
                cls._postgresql = None

    @classmethod
    def get_database_params(cls, psycopg_naming=False):
        if cls._postgresql:
            params = cls._postgresql.dsn()

            if psycopg_naming:
                db_name_1 = params.get("dbname")
                db_name_2 = params.get("database")
                if not db_name_1 and db_name_2:
                    params["dbname"] = db_name_2
                    del params["database"]

            return params
        else:
            return {}

    @classmethod
    def execute_commands(cls, commands: List[str]):
        if not cls._postgresql:
            raise SetupTestException("Database not initialized!")

        with psycopg.connect(**SetupTest.get_database_params(psycopg_naming=True)) as connection:
            with connection.cursor() as cursor:
                for command in commands:
                    try:
                        cursor.execute(command)
                    except Exception as ex:
                        _logger.error("db-command failed: %s\n%s", ex, command)
                        raise

    @classmethod
    def query_one(cls, query: str):
        if not cls._postgresql:
            raise SetupTestException("Database not initialized!")

        with psycopg.connect(**SetupTest.get_database_params(psycopg_naming=True)) as connection:
            with connection.cursor(row_factory=dict_row) as cursor:
                cursor.execute(query)
                return cursor.fetchone()

    @classmethod
    def query_all(cls, query: str):
        if not cls._postgresql:
            raise SetupTestException("Database not initialized!")

        with psycopg.connect(**SetupTest.get_database_params(psycopg_naming=True)) as connection:
            with connection.cursor(row_factory=dict_row) as cursor:
                cursor.execute(query)
                return cursor.fetchall()



import logging
import os
import sys
import logging.handlers


class LoggingConfKey:
    """Predefined keys to be reused in inherited worker classes (section: "worker_settings"). goal: consistent naming"""
    FILE = "file"
    LEVEL = "level"
    MAX_BYTES = "max_bytes"
    MAX_COUNT = "max_count"

    PRINT_CONSOLE = "print_console"
    SKIP_TIMES = "skip_times"

    MODULES_LEVELS = "module_levels"


LOGGING_DEFAULT_LOG_LEVEL = "info"
LOGGING_CHOICES = ["debug", "info", "warning", "error"]


LOGGING_JSONSCHEMA = {
    "type": "object",
    "properties": {
        LoggingConfKey.FILE: {"type": "string", "minLength": 3, "description": "Log file (path)"},
        LoggingConfKey.LEVEL: {"type": "string", "enum": LOGGING_CHOICES, "description": "Log level"},
        LoggingConfKey.MAX_BYTES: {"type": "integer", "minimum": 102400, "description": "Max bytes per log files."},
        LoggingConfKey.MAX_COUNT: {"type": "integer", "minimum": 1, "description": "Max count of rolled log files."},
        LoggingConfKey.PRINT_CONSOLE: {"type": "boolean", "description": "Print logs to console too."},
        LoggingConfKey.SKIP_TIMES: {"type": "boolean", "description": "Skip timestamps in log output and prints to console."},

        LoggingConfKey.MODULES_LEVELS: {
            "type": "object",
            "additionalProperties": {"type": "string", "enum": LOGGING_CHOICES},
            "description": "Dictionary of <module name as shown in log>:<log level>"
        }
    },
    "additionalProperties": False,
}


# noinspection SpellCheckingInspection
class ServiceLogging:

    @classmethod
    def configure(cls, config, file_cli, level_cli, print_console_cli, skip_times_cli):
        handlers = []

        if not file_cli:
            file_cli = config.get(LoggingConfKey.FILE)

        if not level_cli:
            level_cli = config.get(LoggingConfKey.LEVEL)
        level_cli = cls.parse_log_level(level_cli)

        if not print_console_cli:
            print_console_cli = config.get(LoggingConfKey.PRINT_CONSOLE, False)
        if not skip_times_cli:
            skip_times_cli = config.get(LoggingConfKey.SKIP_TIMES, False)

        format_with_ts = '%(asctime)s [%(levelname)8s] %(name)s: %(message)s'
        format_no_ts = '[%(levelname)8s] %(name)s: %(message)s'

        if file_cli:
            log_dir = os.path.dirname(file_cli)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir, exist_ok=True)

            max_bytes = config.get(LoggingConfKey.MAX_BYTES, 1048576)
            max_count = config.get(LoggingConfKey.MAX_COUNT, 5)
            handler = logging.handlers.RotatingFileHandler(
                file_cli,
                maxBytes=int(max_bytes),
                backupCount=int(max_count)
            )
            formatter = logging.Formatter(format_with_ts)
            handler.setFormatter(formatter)
            handlers.append(handler)

        if skip_times_cli:
            log_format = format_no_ts
        else:
            log_format = format_with_ts

        if print_console_cli or skip_times_cli:
            handlers.append(logging.StreamHandler(sys.stdout))

        logging.basicConfig(
            format=log_format,
            level=level_cli,
            handlers=handlers
        )

        module_levels = config.get(LoggingConfKey.MODULES_LEVELS, {})
        for logger_name, log_level in module_levels.items():
            log_level = cls.parse_log_level(log_level)
            logger = logging.getLogger(logger_name)
            logger.setLevel(log_level)

    @classmethod
    def parse_log_level(cls, value):
        value = value or LOGGING_DEFAULT_LOG_LEVEL

        if not isinstance(value, type(logging.INFO)):
            input_value = str(value).lower().strip() if value is not None else value
            if input_value == "debug":
                value = logging.DEBUG
            elif input_value == "info":
                value = logging.INFO
            elif input_value == "warning":
                value = logging.WARNING
            elif input_value == "error":
                value = logging.ERROR
            else:
                value = logging.INFO

        return value

    @classmethod
    def get_log_name(cls, instance_class, instance_name):
        class_name = instance_class.__class__.__name__

        # logname ist used for config handling, which ist all lower case
        if instance_name:
            instance_name = instance_name.lower()

        if instance_name and instance_name != class_name.lower() and instance_name != class_name.lower():
            log_name = f"{instance_class.__class__.__name__}({instance_name})"
        else:
            log_name = class_name
        return log_name

    @classmethod
    def get_full_log_name(cls, instance_class, instance_name):
        class_path = instance_class.__module__
        class_name = instance_class.__class__.__name__
        class_full = "{}.{}".format(class_path, class_name)

        # logname ist used for config handling, which ist all lower case
        if instance_name:
            instance_name = instance_name.lower()

        if instance_name and instance_name != class_name.lower() and instance_name != class_full.lower():
            log_name = f"{class_path}.{instance_class.__class__.__name__}({instance_name})"
        else:
            log_name = class_full
        return log_name



import os


class DatabaseUtils:

    @classmethod
    def load_script_file(cls, file: str) -> str:
        """Loads commands from a file"""

        if not os.path.isfile(file):
            raise FileNotFoundError("Script file ({}) not found".format(file))

        with open(file) as f:
            lines = f.readlines()

        return "".join(lines)



import os
import unittest

from test.setup_test import SetupTest
from worker_bunch.database.database_utils import DatabaseUtils


class TestStep(unittest.TestCase):

    def setUp(self):
        test_dir = SetupTest.ensure_test_dir()
        self.script_file = os.path.join(test_dir, "test_script_file.sql")

        if os.path.exists(self.script_file):
            os.remove(self.script_file)

        self.assertFalse(os.path.exists(self.script_file))

    @classmethod
    def write_script_file(cls, file: str, content: str):
        if os.path.exists(file):
            os.remove(file)

        with open(file, 'w') as f:
            f.write(content)

    # noinspection SpellCheckingInspection
    def test_load_script_file(self):
        text_in = "\nabc \n gäüßhh\n\n"
        self.write_script_file(self.script_file, text_in)

        text_out = DatabaseUtils.load_script_file(self.script_file)
        self.assertEqual(text_out, text_in)

    def test_file_not_exists(self):
        with self.assertRaises(FileNotFoundError) as context:
            DatabaseUtils.load_script_file(self.script_file)

        self.assertIn(self.script_file, str(context.exception))



from typing import Dict

import attr


WORKER_INSTANCES_JSONSCHEMA = {
    "type": "object",
    "additionalProperties": {"type": "string"},
    "description": "Dictionary of <worker name>:<worker class path>"
}


@attr.frozen
class WorkerSettingsDeclaration:

    # cron syntax
    settings_schema: Dict[str, any]

    required: bool



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

# this grabs the requirements from requirements.txt
install_requires = [i.strip() for i in open("requirements.txt").readlines()]

VERSION = '0.8.4'
DESCRIPTION = 'Tasks/jobs/rules engine, primarily intended for use in a smarthome environment.'

# Setting up
setup(
    author="Rosenloecher-IT (Raul Rosenlöcher)",
    author_email="<github@rosenloecher-it.de>",
    description=DESCRIPTION,
    install_requires=install_requires,
    keywords=['smarthome', 'mqtt', 'rule-engine', 'task-engine'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="worker-bunch",
    packages=find_packages(),
    url='https://github.com/rosenloecher-it/worker-bunch',
    version=VERSION,
    classifiers=[  # https://pypi.org/classifiers/
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: MIT License',
    ]
)



import abc
import copy
import logging
import os
import threading
from enum import Enum
from logging import Logger
from typing import Dict, List, Optional, Set

from worker_bunch.dispatcher import Dispatcher, DispatcherListener
from worker_bunch.mqtt.mqtt_proxy import MqttProxy
from worker_bunch.notification import Notification
from worker_bunch.service_config import ConfigException
from worker_bunch.service_logging import ServiceLogging
from worker_bunch.utils.time_utils import TimeUtils


class ShutdownException(Exception):
    pass


class WorkerSetup(Enum):
    """keys used in `Worker.setup` kwargs"""
    ASTRAL_TIME_MANAGER = "astral_time_manager"
    BASE_DATA_DIR = "base_data_dir"
    DATABASE_MANAGER = "database_manager"
    MQTT_PROXY = "mqtt_proxy"
    WORKER_SETTINGS = "worker_settings"


class Worker(threading.Thread, DispatcherListener):

    def __init__(self, name: str):
        threading.Thread.__init__(self, name=name)

        self._setup_done = False

        self._lock = threading.Lock()
        self._closing = False  # shutdown in process
        self._base_data_dir: Optional[str] = None

        self.__logger: Optional[Logger] = None

        self._notifications: Set[Notification] = set()

        self._worker_settings: Dict[str, any] = {}
        self._mqtt_proxy: Optional[MqttProxy] = None

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.name)

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.name)

    def setup(self, props: Dict[WorkerSetup, any]):
        with self._lock:
            if self._setup_done:
                raise RuntimeError(f"Setup only once ('{self.name}')!")
            self._setup_done = True

            self._base_data_dir = props.get(WorkerSetup.BASE_DATA_DIR)
            worker_settings = props.get(WorkerSetup.WORKER_SETTINGS)
            self._worker_settings = copy.deepcopy(worker_settings) if worker_settings else {}
            self._mqtt_proxy = props.get(WorkerSetup.MQTT_PROXY)

    def ensure_data_path(self, file_name: Optional[str] = None) -> str:
        if not self._base_data_dir:
            raise ConfigException(f"No data dir configured (but expected by '{self.name}')!")
        data_path = os.path.join(self._base_data_dir, self.name)  # base dir + worker name
        os.makedirs(data_path, exist_ok=True)
        if file_name:
            data_path = os.path.join(data_path, file_name)
        return data_path

    def get_partial_settings_schema(self) -> Optional[Dict[str, any]]:
        """returns a partial JSON schema for extra setting if needed. overwrite..."""
        return None

    def make_partial_settings_required(self) -> bool:
        """return True to force existence via JSON schema validation. overwrite..."""
        return False

    def set_last_will(self):
        """Set the last will at mqtt client. Explicitly triggered before the mqtt client connects. Kind of:
        ```
        if self._topic and self._last_will:
            self._mqtt_proxy.set_last_will(self._topic, self._last_will)
        ```
        """

    def stop(self):
        """
        Just the notification to finish and stop the thread. A last will may be better send within `_final_work`.
        """
        with self._lock:
            self._closing = True

    def proceed(self) -> bool:
        with self._lock:
            if self._closing:
                raise ShutdownException()
            return True

    @property
    def _logger(self):
        if self.__logger is None:
            log_name = ServiceLogging.get_log_name(self, self.name)
            self.__logger = logging.getLogger(log_name)
        return self.__logger

    # noinspection PyMethodMayBeStatic
    def _sleep(self):
        TimeUtils.sleep(0.5)

    @abc.abstractmethod
    def subscribe_notifications(self, dispatcher: Dispatcher):
        """
        subscribe for notifications via:

        timer_job = schedule.every(14 * 60).to(16 * 60).seconds  # type: schedule.Job
        dispatcher.subscribe_timer(self, timer_job, "keyword-timer")

        self._dispatcher.subscribe_cron("* * * * *", self, "cron-every-minute")

        self._dispatcher.subscribe_topics(self, ["mqtt/topic"],
        """
        raise NotImplementedError()

    def add_notifications(self, notifications: Set[Notification]):
        with self._lock:
            self._notifications |= notifications

    def _get_and_reset_notifications(self) -> List[Notification]:
        with self._lock:
            notifications = list(self._notifications)
            self._notifications.clear()
            return notifications

    def _should_handle_pending_notifications(self) -> bool:
        with self._lock:
            return bool(self._notifications)

    def run(self):
        try:
            while self.proceed():
                self._process_notifications()
                self._sleep()

        except ShutdownException:
            pass
        except Exception as ex:
            self._logger.exception(ex)
        finally:
            self._final_work()

    def run_single(self):
        try:
            if self.proceed():
                self._process_notifications()

        except ShutdownException:
            pass
        except Exception as ex:
            self._logger.exception(ex)
        finally:
            self._final_work()

    def _process_notifications(self):
        if self._should_handle_pending_notifications():
            notifications = self._get_and_reset_notifications()
            if notifications:
                self._work(notifications)

    @abc.abstractmethod
    def _work(self, notifications: List[Notification]):
        """the task "main" function."""
        raise NotImplementedError()

    def _final_work(self):
        """
        E.g.: You may send explicitly the last will(s) here (as a normal message). Kind of:
        ```
        if self._topic and self._last_will:
            self._mqtt_proxy.publish(self._topic, self._last_will)
        ```
        """



import abc
import logging
from typing import Dict, List, Optional, Set

import rx
import attr
import schedule
from rx import operators as rx_ops
from paho.mqtt.client import MQTTMessage
from rx.core import Observer
from rx.disposable import Disposable

from worker_bunch.astral_times.astral_times_manager import AstralTimesManager
from worker_bunch.notification import Notification, NotificationType, NotificationBucket
from worker_bunch.service_config import ConfigException
from worker_bunch.utils.time_utils import TimeUtils


_logger = logging.getLogger(__name__)


class DispatcherListener:

    @abc.abstractmethod
    def add_notifications(self, notifications: Set[Notification]):
        raise NotImplementedError()

    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError()


@attr.frozen
class TopicMatch:

    topic: str
    search_pattern: str = None
    listeners: Set[DispatcherListener] = attr.Factory(set)


class Dispatcher:
    """
    Runs only in one thread context.
    """

    DEFAULT_DEBOUNCE_TIME = 0.1

    def __init__(self, astral_time_manager: AstralTimesManager):

        self._astral_time_manager = astral_time_manager

        self._shutdown = False

        self._last_timer_execution = TimeUtils.now().replace(second=0, microsecond=0)

        self._max_debounce_time = 0

        # separation of notifications and trigger, notifications are overwritten by newer ones
        self._notifications: Dict[DispatcherListener, NotificationBucket] = {}

        self._exact_topic_matches: Dict[str, TopicMatch] = {}
        self._wildcard_topic_matches: List[TopicMatch] = []

        self._astral_subscriptions: Dict[str, TopicMatch] = {}
        self._timer_subscriptions: Set[DispatcherListener] = set()  # only to send SINGLE notifications
        self._cron_subscriptions: Dict[str, TopicMatch] = {}
        self._observers: Dict[DispatcherListener, Optional[Observer]] = {}

        # only simple values can be pushed through pipelines. So an "listener-id" instead if the listener reference gets pushed.
        self._observer_listener: Dict[int, DispatcherListener] = {}

        self._disposables: List[Disposable] = []

    def close(self):
        self._shutdown = True

        for observable in self._observers.values():
            observable.on_completed()
        self._observers = {}

        wait_for_emptied_queues = min(self._max_debounce_time + 0.05, 1.0)
        TimeUtils.sleep(wait_for_emptied_queues)

        for disposable in self._disposables:
            disposable.dispose()
        self._disposables = []

    def get_mqtt_topics(self) -> List[str]:
        topics = [m.topic for m in self._wildcard_topic_matches]
        topics.extend([m.topic for m in self._exact_topic_matches.values()])
        return topics

    @classmethod
    def check_and_extract_wildcard_topic(cls, topic):
        if topic.endswith("#"):
            return topic.strip("#")
        else:
            return None

    def _register_mqtt_topic(self, listener: DispatcherListener, topic: str):
        wildcard_search_pattern = self.check_and_extract_wildcard_topic(topic)
        if wildcard_search_pattern:
            topic_match = next((m for m in self._wildcard_topic_matches if m.topic == topic), None)
            if topic_match is None:
                topic_match = TopicMatch(topic=topic, search_pattern=wildcard_search_pattern)
                self._wildcard_topic_matches.append(topic_match)
            topic_match.listeners.add(listener)

        else:
            topic_match = self._exact_topic_matches.get(topic)
            if topic_match is None:
                topic_match = TopicMatch(topic=topic)
                self._exact_topic_matches[topic] = topic_match
            topic_match.listeners.add(listener)

    def subscribe_mqtt_topics(self, listener: DispatcherListener, topics: List[str],
                              debounce_time: float = DEFAULT_DEBOUNCE_TIME) -> None:
        if listener in self._observer_listener:
            raise RuntimeError(f"Listener ({listener.name}) has already subscribed (only one pipeline per listener)!")
        self._observer_listener[id(listener)] = listener

        self._max_debounce_time = max(self._max_debounce_time, debounce_time)

        for topic in topics:
            self._register_mqtt_topic(listener, topic)

        def creating_observer_callback(observer, _):
            self._observers[listener] = observer

        observable = rx.create(creating_observer_callback)

        instance = self
        disposable = observable.pipe(
            rx_ops.debounce(debounce_time)
        ).subscribe(lambda listener_id: instance._send_notifications_by_id(listener_id))

        self._disposables.append(disposable)

    def subscribe_astral_or_cron(self, listener: DispatcherListener, astral_or_cron: str, topic: str):
        if self._astral_time_manager.is_valid_astral_time_key(astral_or_cron):
            self.subscribe_astral_time(listener, astral_or_cron, topic)
        elif TimeUtils.is_cron_time_syntax(astral_or_cron):
            self.subscribe_cron(listener, astral_or_cron, topic)
        else:
            raise ConfigException(f"No astral nor cron format ('{astral_or_cron}'; worker: {listener.name})!")

    def subscribe_astral_time(self, listener: DispatcherListener, astral_key: str, topic: str):
        if not self._astral_time_manager.is_valid_astral_time_key(astral_key):
            raise ConfigException(f"wrong astral format ('{astral_key}'; worker: {listener.name})!")

        topic_match = self._astral_subscriptions.get(astral_key)
        if topic_match is None:
            topic_match = TopicMatch(topic=topic)
            self._astral_subscriptions[astral_key] = topic_match
        topic_match.listeners.add(listener)

    def subscribe_cron(self, listener: DispatcherListener, cron: str, topic: str):
        try:
            TimeUtils.hits_cron_time(cron)  # result down not matter here, just fail immediately
        except ValueError as ex:
            raise ConfigException(f"wrong cron format (cron: '{cron}'; worker: {listener.name}): {str(ex)}!")

        topic_match = self._cron_subscriptions.get(cron)
        if topic_match is None:
            topic_match = TopicMatch(topic=topic)
            self._cron_subscriptions[cron] = topic_match
        topic_match.listeners.add(listener)

    def subscribe_timer(self, listener: DispatcherListener, timer_job: schedule.Job, topic: str):
        """
        :param listener:
        :param timer_job:
        :param topic:
        :return:
        """
        instance = self

        def timer_closure():
            notification = Notification(
                type=NotificationType.TIMER,
                topic=topic,
                payload=None
            )
            instance._store_notification(listener, notification)
            instance._send_notifications(listener)

        timer_job.do(timer_closure)

        self._timer_subscriptions.add(listener)

    def trigger_timers(self):
        """Triggers timer and cron notification,"""
        if self._shutdown:
            return

        schedule.run_pending()

        now = TimeUtils.now().replace(second=0, microsecond=0)

        if self._last_timer_execution < now:  # next minute
            self._last_timer_execution = now
            send_to: Set[DispatcherListener] = set()

            for cron, topic_match in self._cron_subscriptions.items():
                if TimeUtils.hits_cron_time(cron, now):
                    notification = Notification(type=NotificationType.CRON, topic=topic_match.topic, payload=None)
                    for listener in topic_match.listeners:
                        self._store_notification(listener, notification)
                        send_to.add(listener)

            for astral_key, topic_match in self._astral_subscriptions.items():
                if self._astral_time_manager.hits(astral_key, now):
                    notification = Notification(type=NotificationType.ASTRAL, topic=topic_match.topic, payload=None)
                    for listener in topic_match.listeners:
                        self._store_notification(listener, notification)
                        send_to.add(listener)

            for listener in send_to:
                self._send_notifications(listener)

    def trigger_start_notification(self, listeners: List[DispatcherListener], notification_type: NotificationType = None):
        """Send a message to all workers/listeners, that we just started"""
        if not notification_type:
            notification_type = NotificationType.JUST_STARTED
        notification = Notification(type=notification_type, topic="", payload=None)

        for listener in list(listeners):
            self._store_notification(listener, notification)
            self._send_notifications(listener)

    def push_mqtt_messages(self, messages: List[MQTTMessage]):
        if self._shutdown:
            return

        for message in messages:
            notification = Notification.create_from_mqtt(message)

            listeners: Set[DispatcherListener] = set()

            match = self._exact_topic_matches.get(notification.topic)
            if match:
                listeners.update(match.listeners)

            for match in self._wildcard_topic_matches:
                if notification.topic.startswith(match.search_pattern):
                    listeners.update(match.listeners)

            for listener in list(listeners):
                self._store_notification(listener, notification)
                self._queue_notification(listener)

    def _queue_notification(self, listener: DispatcherListener):
        observer = self._observers[listener]  # must exists
        observer.on_next(id(listener))

    def _store_notification(self, listener: DispatcherListener, notification: Notification):
        bucket: Optional[NotificationBucket] = self._notifications.get(listener)
        if bucket is None:
            bucket = NotificationBucket()
            self._notifications[listener] = bucket
        bucket.add(notification)

    def _send_notifications_by_id(self, listener_id: int):
        listener = self._observer_listener.get(listener_id)
        self._send_notifications(listener)

    def _send_notifications(self, listener: DispatcherListener):
        bucket: Optional[NotificationBucket] = self._notifications.get(listener)
        if bucket:
            listener.add_notifications(bucket.get_set())
            bucket.clear()



import copy
import threading

from worker_bunch.database.database_connector import DatabaseConnector
from worker_bunch.service_config import ConfigException


class DatabaseManager:

    def __init__(self, config):

        self._config = copy.deepcopy(config)
        self._lock = threading.Lock()

    def create(self, context_name: str, connection_key: str) -> DatabaseConnector:
        with self._lock:
            connection_config = self._config.get(connection_key)
            if not connection_config:
                raise ConfigException(f"Unknown database connection ({connection_key}) requested!")

            connection_config = copy.deepcopy(connection_config)
            database = DatabaseConnector(connection_config, context_name, connection_key)
            return database



import datetime
import time

import pycron
from tzlocal import get_localzone


class TimeUtils:

    @classmethod
    def now(cls, no_ms=False) -> datetime.datetime:
        """overwrite/mock in test"""
        now = datetime.datetime.now(tz=get_localzone())
        if no_ms:
            now = now.replace(microsecond=0)
        return now

    @classmethod
    def diff_seconds(cls, reference: datetime.datetime):
        return (cls.now() - reference).total_seconds()

    @classmethod
    def sleep(cls, seconds: float) -> int:
        time.sleep(seconds)
        return seconds

    @classmethod
    def hits_cron_time(cls, cron: str, now: datetime.datetime = None):
        """
        Returns True is the cron is triggering right `now`
        """
        now = now if now is not None else cls.now()
        return pycron.is_now(cron, now)

    @classmethod
    def is_cron_time_syntax(cls, cron: str):
        """
        Checks if the string is a proper cron syntax
        """
        try:
            cls.hits_cron_time(cron)  # result down not matter here
            return True
        except ValueError:
            return False



import logging
import os
from typing import Callable, Dict, List, Optional

import attr
from psycopg.errors import DatabaseError
from psycopg.rows import dict_row

from worker_bunch.database.database_config import DatabaseConfKey, MqttOutputType, DATABASE_WORKER_JSONSCHEMA
from worker_bunch.database.database_connector import DatabaseConnector
from worker_bunch.database.database_manager import DatabaseManager
from worker_bunch.database.database_utils import DatabaseUtils
from worker_bunch.dispatcher import Dispatcher
from worker_bunch.notification import Notification, NT
from worker_bunch.service_config import ConfigException
from worker_bunch.utils.time_utils import TimeUtils
from worker_bunch.worker.worker import Worker, WorkerSetup


@attr.define
class Step:
    """Property names from `DatabaseConfKey` must match!"""

    script_file: str = None
    statement: str = None
    replacements: Dict[str, str] = attr.Factory(dict)

    mqtt_last_will: str = None
    mqtt_output_type: str = None
    mqtt_retain: bool = False
    mqtt_topic: str = None


class DatabaseWorker(Worker):

    CRON_TOPIC = "cron"

    def __init__(self, name: str):
        super().__init__(name)

        self._database_manager: Optional[DatabaseManager] = None
        self._connection_key: Optional[str] = None
        self._cron: Optional[str] = None
        self._steps: List[Step] = []
        self._replacements: Dict[str, str] = {}

    def setup(self, props):
        super().setup(props)

        self._database_manager = props[WorkerSetup.DATABASE_MANAGER]

        self._connection_key = self._worker_settings[DatabaseConfKey.CONNECTION_KEY]
        self._cron = self._worker_settings[DatabaseConfKey.CRON]
        self._replacements = self._worker_settings.get(DatabaseConfKey.REPLACEMENTS, {})

        self._steps = []
        config_steps = self._worker_settings[DatabaseConfKey.STEPS]

        found_config_error = False

        def push_error(message: str):
            nonlocal found_config_error
            self._logger.error("config issue - step" + message)
            found_config_error = True

        for index, config_step in enumerate(config_steps):
            step = self.create_step(config_step)
            self._steps.append(step)
            self.prepare_step(step, index, push_error)

        if found_config_error:
            raise ConfigException("Some configuration issues occurred! See log.")

    def get_partial_settings_schema(self) -> Optional[Dict[str, any]]:
        return DATABASE_WORKER_JSONSCHEMA

    def make_partial_settings_required(self) -> bool:
        return True

    def set_last_will(self):
        for step in self._steps:
            if step.mqtt_topic and step.mqtt_last_will:
                self._mqtt_proxy.set_last_will(step.mqtt_topic, step.mqtt_last_will, step.mqtt_retain)

    def subscribe_notifications(self, dispatcher: Dispatcher):
        dispatcher.subscribe_cron(self, self._cron, self.CRON_TOPIC)

    def _work(self, notifications: List[Notification]):
        if not Notification.find(notifications, NT.CRON) and not Notification.find(notifications, NT.SINGLE_STARTED):
            return

        times_log = ""
        time_sum = 0

        database = self._database_manager.create(self.name, self._connection_key)
        with database:
            for index, step in enumerate(self._steps):
                self.proceed()  # raises ShutdownException

                time_start = TimeUtils.now()
                try:
                    if step.mqtt_output_type == MqttOutputType.JSON:
                        self._query_json(database, step)
                    elif step.mqtt_output_type == MqttOutputType.SCALAR:
                        self._query_scalar(database, step)
                    else:
                        self._execute(database, step)

                    database.connection.commit()

                except DatabaseError as ex:
                    self._logger.error(ex)
                    database.connection.rollback()

                time_diff = TimeUtils.diff_seconds(time_start)
                time_sum += time_diff
                if self._logger.isEnabledFor(logging.DEBUG):
                    output_type = f" ({step.mqtt_output_type})" if step.mqtt_output_type else ""
                    self._logger.debug("[%d]: executed statement%s:\n%s", index, output_type, step.statement)
                if time_diff > 0.2 and 1 < len(self._steps):
                    times_log += f"[{index}]={time_diff:.1f}s; "

            times_log += f"sum={time_sum:.1f}s"
            self._logger.info("took: %s", times_log)

    def _final_work(self):
        for step in self._steps:
            if step.mqtt_topic and step.mqtt_last_will:
                self._mqtt_proxy.queue(step.mqtt_topic, step.mqtt_last_will, step.mqtt_retain)

    def _query_json(self, database: DatabaseConnector, step: Step):
        with database.connection.cursor(row_factory=dict_row) as cursor:
            cursor.execute(step.statement)
            fetched = cursor.fetchone()

            self._mqtt_proxy.queue(step.mqtt_topic, fetched, step.mqtt_retain)

    def _query_scalar(self, database: DatabaseConnector, step: Step):
        with database.connection.cursor(row_factory=dict_row) as cursor:
            cursor.execute(step.statement)
            fetched = cursor.fetchone()

            if len(fetched) != 1:
                raise ConfigException("Scalar statement must return exactly 1 result column!")

            payload = str(list(fetched.values())[0])
            self._mqtt_proxy.queue(step.mqtt_topic, payload, step.mqtt_retain)

    @classmethod
    def _execute(cls, database: DatabaseConnector, step: Step):
        with database.connection.cursor() as cursor:
            cursor.execute(step.statement)

    @classmethod
    def create_step(cls, config_step: Dict[str, any]) -> Step:
        return Step(**config_step)

    def prepare_step(self, step: Step, index: int, push_error: Callable):
        if step.mqtt_output_type == MqttOutputType.NONE or not step.mqtt_output_type:
            step.mqtt_output_type = None
        step.mqtt_retain = bool(step.mqtt_retain)

        if not step.mqtt_topic and (step.mqtt_last_will or step.mqtt_output_type):
            push_error(f"[{index}]: missing mqtt topic!")

        self.prepare_step_statement(step, index, push_error)

    def prepare_step_statement(self, step: Step, index: int, push_error: Callable):
        if not step.statement:  # must be from file
            if not os.path.isfile(step.script_file):
                push_error(f"[{index}]: script file ({step.script_file}) not found!")
                return

            try:
                step.statement = DatabaseUtils.load_script_file(step.script_file)
            except IOError as ex:
                push_error(f"[{index}]: script file ({step.script_file}) cannot be read! {str(ex)}")

        replacements = {**self._replacements, **step.replacements}
        for pattern, replacement in replacements.items():
            step.statement = step.statement.replace(pattern, replacement)

        step.statement = step.statement.strip()
        if not step.statement:
            push_error(f"[{index}]: script file ({step.script_file}) is empty!")
            return
